@@ -47,9 +47,9 @@ public class AssertTest {
   }
 
   @Test(description = "GITHUB-2483", expectedExceptions = AssertionError.class)
-  public void testAsymetricNotEquals() {
-    AsymetricEquals equalsSame = AsymetricEquals.equalsSame();
-    AsymetricEquals equalsAll = AsymetricEquals.equalsAll();
+  public void testAsymmetricNotEquals() {
+    AsymmetricEquals equalsSame = AsymmetricEquals.equalsSame();
+    AsymmetricEquals equalsAll = AsymmetricEquals.equalsAll();
     // sanity
     Assert.assertFalse(equalsSame.equals(equalsAll));
     Assert.assertTrue(equalsAll.equals(equalsSame));
@@ -305,22 +305,10 @@ public class AssertTest {
     Assert.assertEquals(actual, expected, 0.1d);
   }
 
-  @SuppressWarnings("serial")
   @Test(expectedExceptions = AssertionError.class)
   public void assertEqualsMapShouldFail() {
-    Map<String, String> mapActual =
-        new HashMap<String, String>() {
-          {
-            put("a", "1");
-          }
-        };
-    Map<String, String> mapExpected =
-        new HashMap<String, String>() {
-          {
-            put("a", "1");
-            put("b", "2");
-          }
-        };
+    Map<String, String> mapActual = Map.of("a", "1");
+    Map<String, String> mapExpected = Map.of("a", "1", "b", "2");
 
     Assert.assertEquals(mapActual, mapExpected);
   }
@@ -367,6 +355,33 @@ public class AssertTest {
     Assert.assertEquals("x", "y");
   }
 
+  @Test(
+      description = "GITHUB-3135",
+      expectedExceptions = AssertionError.class,
+      expectedExceptionsMessageRegExp =
+          "Arrays differ at element \\[2\\]: 3 != 5 expected \\[3\\] but found \\[5\\]")
+  public void testAssertEqualsArrayElementDiffers() {
+    Assert.assertEquals(new Integer[] {1, 2, 5}, new Integer[] {1, 2, 3});
+  }
+
+  @Test(
+      description = "GITHUB-3135",
+      expectedExceptions = AssertionError.class,
+      expectedExceptionsMessageRegExp =
+          "Arrays differ at element \\[2\\]: 3 != null expected \\[3\\] but found \\[null\\]")
+  public void testAssertEqualsArrayElementIsNull() {
+    Assert.assertEquals(new Integer[] {1, 2, null}, new Integer[] {1, 2, 3});
+  }
+
+  @Test(
+      description = "GITHUB-3135",
+      expectedExceptions = AssertionError.class,
+      expectedExceptionsMessageRegExp =
+          "Arrays differ at element \\[2\\]: null != 5 expected \\[null\\] but found \\[5\\]")
+  public void testAssertEqualsArrayElementIsNotNull() {
+    Assert.assertEquals(new Integer[] {1, 2, 5}, new Integer[] {1, 2, null});
+  }
+
   @Test
   public void testAssertEqualsNoOrder() {
     String[] actual = {"a", "b"};
@@ -565,7 +580,29 @@ public class AssertTest {
     Assert.assertNotEquals(obj, obj);
   }
 
-  class Contrived {
+  @Test(description = "GITHUB-3140")
+  public void testAssertEqualsDeepSet() {
+    var expectedSet = new LinkedHashSet<>();
+    expectedSet.add(new Contrived(1));
+    expectedSet.add(new Contrived[] {new Contrived(1)});
+    var actualSet = new LinkedHashSet<>();
+    actualSet.add(new Contrived(1));
+    actualSet.add(new Contrived[] {new Contrived(1)});
+    Assert.assertEqualsDeep(actualSet, expectedSet);
+  }
+
+  @Test(description = "GITHUB-3140", expectedExceptions = AssertionError.class)
+  public void testAssertEqualsDeepSetFail() {
+    var expectedSet = new LinkedHashSet<>();
+    expectedSet.add(new Contrived(1));
+    expectedSet.add(new Contrived[] {new Contrived(1)});
+    var actualSet = new LinkedHashSet<>();
+    actualSet.add(new Contrived(1));
+    actualSet.add(new Contrived[] {new Contrived(2)});
+    Assert.assertEqualsDeep(actualSet, expectedSet);
+  }
+
+  static class Contrived {
 
     int integer;
 
@@ -617,28 +654,28 @@ public class AssertTest {
     }
   }
 
-  static class AsymetricEquals {
+  static class AsymmetricEquals {
 
-    static AsymetricEquals equalsAll() {
-      return new AsymetricEquals(null);
+    static AsymmetricEquals equalsAll() {
+      return new AsymmetricEquals(null);
     }
 
-    static AsymetricEquals equalsSame() {
-      return new AsymetricEquals(new Object());
+    static AsymmetricEquals equalsSame() {
+      return new AsymmetricEquals(new Object());
     }
 
     Object value;
 
-    private AsymetricEquals(Object value) {
+    private AsymmetricEquals(Object value) {
       this.value = value;
     }
 
     @Override
     public boolean equals(Object o) {
       if (this == o) return true;
-      if (!(o instanceof AsymetricEquals)) return false;
+      if (!(o instanceof AsymmetricEquals)) return false;
 
-      AsymetricEquals that = (AsymetricEquals) o;
+      AsymmetricEquals that = (AsymmetricEquals) o;
       if (value == null) return true;
       return value.equals(that.value);
     }
@@ -654,6 +691,11 @@ public class AssertTest {
     public boolean equals(Object o) {
       return true; // broken implementation
     }
+
+    @Override
+    public int hashCode() {
+      return 0;
+    }
   }
 
   static class BrokenEqualsFalse {
@@ -661,5 +703,10 @@ public class AssertTest {
     public boolean equals(Object o) {
       return false; // broken implementation
     }
+
+    @Override
+    public int hashCode() {
+      return 0;
+    }
   }
 }
