@@ -10,11 +10,11 @@ java {
 
 dependencies {
     api(projects.testngCollections)
-    api("com.google.code.findbugs:jsr305:_")
-    "guiceApi"(platform("com.google.inject:guice-bom:_"))
+    compileOnly("com.github.spotbugs:spotbugs:4.8.1")
+    "guiceApi"(platform("com.google.inject:guice-bom:5.1.0"))
     "guiceApi"("com.google.inject:guice")
 
     implementation(projects.testngReflectionUtils)
-    api("org.slf4j:slf4j-api:_")
-    testImplementation("org.slf4j:slf4j-simple:_")
+    api("org.slf4j:slf4j-api:2.0.16")
+    testImplementation("org.slf4j:slf4j-simple:2.0.16")
 }
