@@ -18,6 +18,7 @@ import org.testng.internal.invokers.ConfigMethodArguments.Builder;
 import org.testng.internal.invokers.IInvocationStatus;
 import org.testng.internal.invokers.IInvoker;
 import org.testng.internal.invokers.InvokedMethod;
+import org.testng.internal.objects.ObjectFactoryImpl;
 import org.testng.internal.thread.ThreadUtil;
 import org.testng.reporters.JUnitXMLReporter;
 import org.testng.reporters.TestHTMLReporter;
@@ -29,25 +30,24 @@ import org.testng.xml.XmlTest;
  * <CODE>SuiteRunner</CODE> is responsible for running all the tests included in one suite. The test
  * start is triggered by {@link #run()} method.
  */
-public class SuiteRunner implements ISuite, IInvokedMethodListener {
+public class SuiteRunner implements ISuite, ISuiteRunnerListener {
 
   private static final String DEFAULT_OUTPUT_DIR = "test-output";
 
-  private final Map<String, ISuiteResult> suiteResults =
-      Collections.synchronizedMap(Maps.newLinkedHashMap());
+  private final Map<String, ISuiteResult> suiteResults = Maps.newLinkedHashMap();
   private final List<TestRunner> testRunners = Lists.newArrayList();
   private final Map<Class<? extends ISuiteListener>, ISuiteListener> listeners =
       Maps.newLinkedHashMap();
 
   private String outputDir;
-  private XmlSuite xmlSuite;
+  private final XmlSuite xmlSuite;
   private Injector parentInjector;
 
   private final List<ITestListener> testListeners = Lists.newArrayList();
   private final Map<Class<? extends IClassListener>, IClassListener> classListeners =
       Maps.newLinkedHashMap();
-  private ITestRunnerFactory tmpRunnerFactory;
-  private final DataProviderHolder holder = new DataProviderHolder();
+  private final ITestRunnerFactory tmpRunnerFactory;
+  private final DataProviderHolder holder;
 
   private boolean useDefaultListeners = true;
 
@@ -56,18 +56,19 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
 
   // The configuration
   // Note: adjust test.multiplelisteners.SimpleReporter#generateReport test if renaming the field
-  private IConfiguration configuration;
+  private final IConfiguration configuration;
 
   private ITestObjectFactory objectFactory;
   private Boolean skipFailedInvocationCounts = Boolean.FALSE;
   private final List<IReporter> reporters = Lists.newArrayList();
 
-  private Map<Class<? extends IInvokedMethodListener>, IInvokedMethodListener>
+  private final Map<Class<? extends IInvokedMethodListener>, IInvokedMethodListener>
       invokedMethodListeners;
 
   private final SuiteRunState suiteState = new SuiteRunState();
   private final IAttributes attributes = new Attributes();
   private final Set<IExecutionVisualiser> visualisers = Sets.newHashSet();
+  private final ITestListener exitCodeListener;
 
   public SuiteRunner(
       IConfiguration configuration,
@@ -93,9 +94,9 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
         useDefaultListeners,
         new ArrayList<>() /* method interceptor */,
         null /* invoked method listeners */,
-        null /* test listeners */,
+        new TestListenersContainer() /* test listeners */,
         null /* class listeners */,
-        new DataProviderHolder(),
+        new DataProviderHolder(configuration),
         comparator);
   }
 
@@ -106,48 +107,26 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
       ITestRunnerFactory runnerFactory,
       boolean useDefaultListeners,
       List<IMethodInterceptor> methodInterceptors,
-      Collection<IInvokedMethodListener> invokedMethodListeners,
-      Collection<ITestListener> testListeners,
+      Collection<IInvokedMethodListener> invokedMethodListener,
+      TestListenersContainer container,
       Collection<IClassListener> classListeners,
       DataProviderHolder holder,
       Comparator<ITestNGMethod> comparator) {
-    init(
-        configuration,
-        suite,
-        outputDir,
-        runnerFactory,
-        useDefaultListeners,
-        methodInterceptors,
-        invokedMethodListeners,
-        testListeners,
-        classListeners,
-        holder,
-        comparator);
-  }
-
-  private void init(
-      IConfiguration configuration,
-      XmlSuite suite,
-      String outputDir,
-      ITestRunnerFactory runnerFactory,
-      boolean useDefaultListeners,
-      List<IMethodInterceptor> methodInterceptors,
-      Collection<IInvokedMethodListener> invokedMethodListener,
-      Collection<ITestListener> testListeners,
-      Collection<IClassListener> classListeners,
-      DataProviderHolder attribs,
-      Comparator<ITestNGMethod> comparator) {
     if (comparator == null) {
       throw new IllegalArgumentException("comparator must not be null");
     }
-    this.holder.merge(attribs);
+    this.holder = holder;
     this.configuration = configuration;
     this.xmlSuite = suite;
     this.useDefaultListeners = useDefaultListeners;
     this.tmpRunnerFactory = runnerFactory;
+    this.exitCodeListener = container.exitCodeListener;
     List<IMethodInterceptor> localMethodInterceptors =
         Optional.ofNullable(methodInterceptors).orElse(Lists.newArrayList());
     setOutputDir(outputDir);
+    if (configuration.getObjectFactory() == null) {
+      configuration.setObjectFactory(new ObjectFactoryImpl());
+    }
     if (suite.getObjectFactoryClass() == null) {
       objectFactory = configuration.getObjectFactory();
     } else {
@@ -200,12 +179,9 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
         Optional.ofNullable(invokedMethodListener).orElse(Collections.emptyList())) {
       invokedMethodListeners.put(listener.getClass(), listener);
     }
-    invokedMethodListeners.put(getClass(), this);
 
     skipFailedInvocationCounts = suite.skipFailedInvocationCounts();
-    if (null != testListeners) {
-      this.testListeners.addAll(testListeners);
-    }
+    this.testListeners.addAll(container.listeners);
     for (IClassListener classListener :
         Optional.ofNullable(classListeners).orElse(Collections.emptyList())) {
       this.classListeners.put(classListener.getClass(), classListener);
@@ -254,13 +230,21 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
     useDefaultListeners = reportResults;
   }
 
+  public ITestListener getExitCodeListener() {
+    return exitCodeListener;
+  }
+
   private void invokeListeners(boolean start) {
     if (start) {
-      for (ISuiteListener sl : Lists.newArrayList(listeners.values())) {
+      for (ISuiteListener sl :
+          ListenerOrderDeterminer.order(
+              listeners.values(), this.configuration.getListenerComparator())) {
         sl.onStart(this);
       }
     } else {
-      List<ISuiteListener> suiteListenersReversed = Lists.newReversedArrayList(listeners.values());
+      List<ISuiteListener> suiteListenersReversed =
+          ListenerOrderDeterminer.reversedOrder(
+              listeners.values(), this.configuration.getListenerComparator());
       for (ISuiteListener sl : suiteListenersReversed) {
         sl.onFinish(this);
       }
@@ -285,10 +269,12 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
               testListeners.toArray(new ITestListener[0]),
               useDefaultListeners,
               skipFailedInvocationCounts,
-              comparator);
+              comparator,
+              this);
     } else {
       factory =
-          new ProxyTestRunnerFactory(testListeners.toArray(new ITestListener[0]), tmpRunnerFactory);
+          new ProxyTestRunnerFactory(
+              testListeners.toArray(new ITestListener[0]), tmpRunnerFactory, configuration);
     }
 
     return factory;
@@ -343,6 +329,10 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
       // -- cbeust
       invoker = tr.getInvoker();
 
+      // Add back the configuration listeners that may have gotten altered after
+      // our suite level listeners were invoked.
+      this.configuration.getConfigurationListeners().forEach(tr::addConfigurationListener);
+
       for (ITestNGMethod m : tr.getBeforeSuiteMethods()) {
         beforeSuiteMethods.put(m.getConstructorOrMethod().getMethod(), m);
       }
@@ -424,12 +414,14 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
     }
   }
 
+  private final AutoCloseableLock suiteResultsLock = new AutoCloseableLock();
+
   private void runTest(TestRunner tr) {
     visualisers.forEach(tr::addListener);
     tr.run();
 
     ISuiteResult sr = new SuiteResult(xmlSuite, tr);
-    synchronized (suiteResults) {
+    try (AutoCloseableLock ignore = suiteResultsLock.lock()) {
       suiteResults.put(tr.getName(), sr);
     }
   }
@@ -448,7 +440,11 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
     }
 
     ThreadUtil.execute(
-        "tests", tasks, xmlSuite.getThreadCount(), xmlSuite.getTimeOut(XmlTest.DEFAULT_TIMEOUT_MS));
+        configuration,
+        "tests",
+        tasks,
+        xmlSuite.getThreadCount(),
+        xmlSuite.getTimeOut(XmlTest.DEFAULT_TIMEOUT_MS));
   }
 
   private class SuiteWorker implements Runnable {
@@ -517,7 +513,9 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
 
   @Override
   public Map<String, ISuiteResult> getResults() {
-    return suiteResults;
+    // Just to ensure that we guard the internals of the suite results we now wrap it
+    // around with an unmodifiable map.
+    return Collections.unmodifiableMap(suiteResults);
   }
 
   /**
@@ -580,18 +578,21 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
     private final boolean skipFailedInvocationCounts;
     private final IConfiguration configuration;
     private final Comparator<ITestNGMethod> comparator;
+    private final SuiteRunner suiteRunner;
 
     public DefaultTestRunnerFactory(
         IConfiguration configuration,
         ITestListener[] failureListeners,
         boolean useDefaultListeners,
         boolean skipFailedInvocationCounts,
-        Comparator<ITestNGMethod> comparator) {
+        Comparator<ITestNGMethod> comparator,
+        SuiteRunner suiteRunner) {
       this.configuration = configuration;
       this.failureGenerators = failureListeners;
       this.useDefaultListeners = useDefaultListeners;
       this.skipFailedInvocationCounts = skipFailedInvocationCounts;
       this.comparator = comparator;
+      this.suiteRunner = suiteRunner;
     }
 
     @Override
@@ -610,7 +611,7 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
         Collection<IInvokedMethodListener> listeners,
         List<IClassListener> classListeners,
         Map<Class<? extends IDataProviderListener>, IDataProviderListener> dataProviderListeners) {
-      DataProviderHolder holder = new DataProviderHolder();
+      DataProviderHolder holder = new DataProviderHolder(this.configuration);
       holder.addListeners(dataProviderListeners.values());
       return newTestRunner(suite, test, listeners, classListeners, holder);
     }
@@ -637,7 +638,8 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
               listeners,
               classListeners,
               comparator,
-              holder);
+              holder,
+              suiteRunner);
 
       if (useDefaultListeners) {
         testRunner.addListener(new TestHTMLReporter());
@@ -666,9 +668,13 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
     private final ITestListener[] failureGenerators;
     private final ITestRunnerFactory target;
 
-    public ProxyTestRunnerFactory(ITestListener[] failureListeners, ITestRunnerFactory target) {
+    private final IConfiguration configuration;
+
+    public ProxyTestRunnerFactory(
+        ITestListener[] failureListeners, ITestRunnerFactory target, IConfiguration configuration) {
       failureGenerators = failureListeners;
       this.target = target;
+      this.configuration = configuration;
     }
 
     @Override
@@ -687,7 +693,7 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
         Collection<IInvokedMethodListener> listeners,
         List<IClassListener> classListeners,
         Map<Class<? extends IDataProviderListener>, IDataProviderListener> dataProviderListeners) {
-      DataProviderHolder holder = new DataProviderHolder();
+      DataProviderHolder holder = new DataProviderHolder(configuration);
       holder.addListeners(dataProviderListeners.values());
       return newTestRunner(suite, test, listeners, classListeners, holder);
     }
@@ -801,4 +807,19 @@ public class SuiteRunner implements ISuite, IInvokedMethodListener {
         .flatMap(tr -> Arrays.stream(tr.getAllTestMethods()))
         .collect(Collectors.toList());
   }
+
+  static class TestListenersContainer {
+    private final List<ITestListener> listeners = Lists.newArrayList();
+    private final ITestListener exitCodeListener;
+
+    TestListenersContainer() {
+      this(Collections.emptyList(), null);
+    }
+
+    TestListenersContainer(List<ITestListener> listeners, ITestListener exitCodeListener) {
+      this.listeners.addAll(listeners);
+      this.exitCodeListener =
+          Objects.requireNonNullElseGet(exitCodeListener, () -> new ITestListener() {});
+    }
+  }
 }
