@@ -1,5 +1,6 @@
 package org.testng;
 
+import static org.testng.ListenerComparator.sort;
 import static org.testng.internal.Utils.defaultIfStringEmpty;
 import static org.testng.internal.Utils.isStringEmpty;
 import static org.testng.internal.Utils.isStringNotEmpty;
@@ -15,23 +16,28 @@ import java.util.Collection;
 import java.util.Collections;
 import java.util.List;
 import java.util.Map;
+import java.util.Objects;
+import java.util.Optional;
 import java.util.ServiceLoader;
 import java.util.Set;
 import java.util.concurrent.LinkedBlockingQueue;
-import java.util.concurrent.TimeUnit;
+import org.testng.SuiteRunner.TestListenersContainer;
 import org.testng.annotations.ITestAnnotation;
 import org.testng.collections.Lists;
 import org.testng.collections.Maps;
 import org.testng.collections.Sets;
 import org.testng.internal.ClassHelper;
 import org.testng.internal.Configuration;
+import org.testng.internal.DefaultListenerFactory;
 import org.testng.internal.DynamicGraph;
 import org.testng.internal.ExitCode;
 import org.testng.internal.IConfiguration;
+import org.testng.internal.ListenerOrderDeterminer;
+import org.testng.internal.MethodSorting;
+import org.testng.internal.ObjectBag;
 import org.testng.internal.OverrideProcessor;
 import org.testng.internal.ReporterConfig;
 import org.testng.internal.RuntimeBehavior;
-import org.testng.internal.Systematiser;
 import org.testng.internal.Utils;
 import org.testng.internal.Version;
 import org.testng.internal.annotations.DefaultAnnotationTransformer;
@@ -45,18 +51,14 @@ import org.testng.internal.objects.IObjectDispenser;
 import org.testng.internal.objects.pojo.BasicAttributes;
 import org.testng.internal.objects.pojo.CreationAttributes;
 import org.testng.internal.thread.graph.SuiteWorkerFactory;
-import org.testng.junit.JUnitTestFinder;
 import org.testng.log4testng.Logger;
-import org.testng.reporters.EmailableReporter;
 import org.testng.reporters.EmailableReporter2;
 import org.testng.reporters.FailedReporter;
 import org.testng.reporters.JUnitReportReporter;
-import org.testng.reporters.SuiteHTMLReporter;
+import org.testng.reporters.PerSuiteXMLReporter;
 import org.testng.reporters.VerboseReporter;
 import org.testng.reporters.XMLReporter;
 import org.testng.reporters.jq.Main;
-import org.testng.thread.IExecutorFactory;
-import org.testng.thread.ITestNGThreadPoolExecutor;
 import org.testng.thread.IThreadWorkerFactory;
 import org.testng.util.Strings;
 import org.testng.xml.IPostProcessor;
@@ -128,12 +130,8 @@ public class TestNG {
   protected List<XmlSuite> m_suites = Lists.newArrayList();
   private List<XmlSuite> m_cmdlineSuites;
   private String m_outputDir = DEFAULT_OUTPUTDIR;
-
   private String[] m_includedGroups;
   private String[] m_excludedGroups;
-
-  private Boolean m_isJUnit = XmlSuite.DEFAULT_JUNIT;
-  private Boolean m_isMixed = XmlSuite.DEFAULT_MIXED;
   protected boolean m_useDefaultListeners = true;
   private boolean m_failIfAllTestsSkipped = false;
   private final List<String> m_listenersToSkipFromBeingWiredIn = new ArrayList<>();
@@ -153,8 +151,6 @@ public class TestNG {
   private final Map<Class<? extends IDataProviderInterceptor>, IDataProviderInterceptor>
       m_dataProviderInterceptors = Maps.newLinkedHashMap();
 
-  private IExecutorFactory m_executorFactory = null;
-
   public static final Integer DEFAULT_VERBOSE = 1;
 
   // Command line suite parameters
@@ -276,6 +272,14 @@ public class TestNG {
     m_useDefaultListeners = useDefaultListeners;
   }
 
+  public void setListenerComparator(ListenerComparator listenerComparator) {
+    this.m_configuration.setListenerComparator(listenerComparator);
+  }
+
+  public ListenerComparator getListenerComparator() {
+    return m_configuration.getListenerComparator();
+  }
+
   /**
    * Sets a jar containing a testng.xml file.
    *
@@ -350,11 +354,9 @@ public class TestNG {
         continue;
       }
       // If test names were specified, only run these test names
-      TestNamesMatcher testNamesMatcher = new TestNamesMatcher(s, m_testNames);
-      List<String> missMatchedTestname = testNamesMatcher.getMissMatchedTestNames();
-      if (!missMatchedTestname.isEmpty()) {
-        throw new TestNGException("The test(s) <" + missMatchedTestname + "> cannot be found.");
-      }
+      TestNamesMatcher testNamesMatcher =
+          new TestNamesMatcher(s, m_testNames, m_ignoreMissedTestNames);
+      testNamesMatcher.validateMissMatchedTestNames();
       result.addAll(testNamesMatcher.getSuitesMatchingTestNames());
     }
 
@@ -417,10 +419,12 @@ public class TestNG {
     File jarFile = new File(m_jarPath);
 
     JarFileUtils utils =
-        new JarFileUtils(getProcessor(), m_xmlPathInJar, m_testNames, m_parallelMode);
+        new JarFileUtils(
+            getProcessor(), m_xmlPathInJar, m_testNames, m_parallelMode, m_ignoreMissedTestNames);
 
     Collection<XmlSuite> allSuites = utils.extractSuitesFrom(jarFile);
-    m_suites.addAll(processCommandLineArgs(allSuites));
+    allSuites.forEach(this::processParallelModeCommandLineArgs);
+    m_suites.addAll(allSuites);
   }
 
   /** @param threadCount Define the number of threads in the thread pool. */
@@ -544,11 +548,6 @@ public class TestNG {
       if (test != null) {
         suiteName = defaultIfStringEmpty(test.getSuiteName(), suiteName);
         testName = defaultIfStringEmpty(test.getTestName(), testName);
-      } else {
-        if (m_isMixed && JUnitTestFinder.isJUnitTest(c)) {
-          isJUnit = true;
-          testName = c.getName();
-        }
       }
       XmlSuite xmlSuite = suites.get(suiteName);
       if (xmlSuite == null) {
@@ -571,7 +570,6 @@ public class TestNG {
       if (xmlTest == null) {
         xmlTest = new XmlTest(xmlSuite);
         xmlTest.setName(testName);
-        xmlTest.setJUnit(isJUnit);
       }
 
       xmlTest.getXmlClasses().add(xmlClasses[i]);
@@ -598,6 +596,30 @@ public class TestNG {
     return this.m_configuration.getReportAllDataDrivenTestsAsSkipped();
   }
 
+  public void propagateDataProviderFailureAsTestFailure() {
+    this.m_configuration.propagateDataProviderFailureAsTestFailure();
+  }
+
+  public boolean isPropagateDataProviderFailureAsTestFailure() {
+    return this.m_configuration.isPropagateDataProviderFailureAsTestFailure();
+  }
+
+  public void shareThreadPoolForDataProviders(boolean flag) {
+    this.m_configuration.shareThreadPoolForDataProviders(flag);
+  }
+
+  public boolean isShareThreadPoolForDataProviders() {
+    return this.m_configuration.isShareThreadPoolForDataProviders();
+  }
+
+  public boolean useGlobalThreadPool() {
+    return this.m_configuration.useGlobalThreadPool();
+  }
+
+  public void shouldUseGlobalThreadPool(boolean flag) {
+    this.m_configuration.shouldUseGlobalThreadPool(flag);
+  }
+
   /**
    * Set the suites file names to be run by this TestNG object. This method tries to load and parse
    * the specified TestNG suite xml files. If a file is missing, it is ignored.
@@ -668,8 +690,12 @@ public class TestNG {
    *     IReporter
    */
   public void setListenerClasses(List<Class<? extends ITestNGListener>> classes) {
+    ITestNGListenerFactory factory = m_configuration.getListenerFactory();
+    if (factory == null) {
+      factory = new DefaultListenerFactory(m_objectFactory, null);
+    }
     for (Class<? extends ITestNGListener> cls : classes) {
-      addListener(m_objectFactory.newInstance(cls));
+      addListener(factory.createListener(cls));
     }
   }
 
@@ -705,6 +731,9 @@ public class TestNG {
     if (listener == null) {
       return;
     }
+    if (!listener.isEnabled()) {
+      return;
+    }
     if (listener instanceof IExecutionVisualiser) {
       IExecutionVisualiser visualiser = (IExecutionVisualiser) listener;
       maybeAddListener(m_executionVisualisers, visualiser);
@@ -790,6 +819,8 @@ public class TestNG {
   /** The list of test names to run from the given suite */
   private List<String> m_testNames;
 
+  private boolean m_ignoreMissedTestNames;
+
   private Integer m_suiteThreadPoolSize = CommandLineArgs.SUITE_THREAD_POOL_SIZE_DEFAULT;
 
   private boolean m_randomizeSuites = Boolean.FALSE;
@@ -798,6 +829,7 @@ public class TestNG {
 
   private Boolean m_preserveOrder = XmlSuite.DEFAULT_PRESERVE_ORDER;
   private Boolean m_groupByInstances;
+  private boolean m_generateResultsPerSuite = false;
 
   private IConfiguration m_configuration;
 
@@ -812,29 +844,17 @@ public class TestNG {
     m_verbose = verbose;
   }
 
-  public void setExecutorFactoryClass(String clazzName) {
-    this.m_executorFactory = createExecutorFactoryInstanceUsing(clazzName);
+  public void setExecutorServiceFactory(IExecutorServiceFactory factory) {
+    m_configuration.setExecutorServiceFactory(
+        Objects.requireNonNull(factory, "ExecutorServiceFactory cannot be null"));
   }
 
-  private IExecutorFactory createExecutorFactoryInstanceUsing(String clazzName) {
-    Class<?> cls = ClassHelper.forName(clazzName);
-    Object instance = m_objectFactory.newInstance(cls);
-    if (instance instanceof IExecutorFactory) {
-      return (IExecutorFactory) instance;
-    }
-    throw new IllegalArgumentException(
-        clazzName + " does not implement " + IExecutorFactory.class.getName());
+  public void setListenerFactory(ITestNGListenerFactory factory) {
+    this.m_configuration.setListenerFactory(factory);
   }
 
-  public void setExecutorFactory(IExecutorFactory factory) {
-    this.m_executorFactory = factory;
-  }
-
-  public IExecutorFactory getExecutorFactory() {
-    if (this.m_executorFactory == null) {
-      this.m_executorFactory = createExecutorFactoryInstanceUsing(DEFAULT_THREADPOOL_FACTORY);
-    }
-    return this.m_executorFactory;
+  public void setGenerateResultsPerSuite(boolean generateResultsPerSuite) {
+    this.m_generateResultsPerSuite = generateResultsPerSuite;
   }
 
   private void initializeCommandLineSuites() {
@@ -917,15 +937,15 @@ public class TestNG {
     if (m_failIfAllTestsSkipped) {
       this.exitCodeListener.failIfAllTestsSkipped();
     }
-    addListener(this.exitCodeListener);
     if (m_useDefaultListeners) {
-      addReporter(SuiteHTMLReporter.class);
       addReporter(Main.class);
       addReporter(FailedReporter.class);
-      addReporter(XMLReporter.class);
-      if (RuntimeBehavior.useOldTestNGEmailableReporter()) {
-        addReporter(EmailableReporter.class);
-      } else if (RuntimeBehavior.useEmailableReporter()) {
+      if (m_generateResultsPerSuite) {
+        addReporter(PerSuiteXMLReporter.class);
+      } else {
+        addReporter(XMLReporter.class);
+      }
+      if (RuntimeBehavior.useEmailableReporter()) {
         addReporter(EmailableReporter2.class);
       }
       addReporter(JUnitReportReporter.class);
@@ -973,7 +993,6 @@ public class TestNG {
     m_configuration.setConfigurable(m_configurable);
     m_configuration.setObjectFactory(factory);
     m_configuration.setAlwaysRunListeners(this.m_alwaysRun);
-    m_configuration.setExecutorFactory(getExecutorFactory());
   }
 
   private void addListeners(XmlSuite s) {
@@ -1062,6 +1081,7 @@ public class TestNG {
     m_end = System.currentTimeMillis();
 
     if (null != suiteRunners) {
+      suiteRunners.forEach(ObjectBag::cleanup);
       generateReports(suiteRunners);
     }
 
@@ -1093,23 +1113,32 @@ public class TestNG {
   }
 
   private void runSuiteAlterationListeners() {
-    for (IAlterSuiteListener l : m_alterSuiteListeners.values()) {
+    Collection<IAlterSuiteListener> original =
+        sort(m_alterSuiteListeners.values(), m_configuration.getListenerComparator());
+    for (IAlterSuiteListener l : original) {
       l.alter(m_suites);
     }
   }
 
   private void runExecutionListeners(boolean start) {
-    List<IExecutionListener> executionListeners = m_configuration.getExecutionListeners();
+    List<IExecutionListener> executionListeners =
+        ListenerOrderDeterminer.order(
+            m_configuration.getExecutionListeners(), m_configuration.getListenerComparator());
     if (start) {
       for (IExecutionListener l : executionListeners) {
         l.onExecutionStart();
       }
+      // Invoke our exit code listener after all the user's listeners have run.
+      exitCodeListener.onExecutionStart();
     } else {
       List<IExecutionListener> executionListenersReversed =
-          Lists.newReversedArrayList(executionListeners);
+          ListenerOrderDeterminer.reversedOrder(
+              m_configuration.getExecutionListeners(), m_configuration.getListenerComparator());
       for (IExecutionListener l : executionListenersReversed) {
         l.onExecutionFinish();
       }
+      // Invoke our exit code listener after all the user's listeners have run.
+      exitCodeListener.onExecutionFinish();
     }
   }
 
@@ -1121,7 +1150,11 @@ public class TestNG {
   }
 
   private void generateReports(List<ISuite> suiteRunners) {
-    for (IReporter reporter : m_reporters.values()) {
+    List<IReporter> reporters = new ArrayList<>(m_reporters.values());
+    // Add our Exit code listener as the last of the reporter so that we can still accommodate
+    // whatever changes were done by a user's reporting listener
+    reporters.add(exitCodeListener);
+    for (IReporter reporter : reporters) {
       try {
         long start = System.currentTimeMillis();
         reporter.generateReport(m_suites, suiteRunners, m_outputDir);
@@ -1157,6 +1190,12 @@ public class TestNG {
     // First initialize the suite runners to ensure there are no configuration issues.
     // Create a map with XmlSuite as key and corresponding SuiteRunner as value
     for (XmlSuite xmlSuite : m_suites) {
+      if (m_configuration.isShareThreadPoolForDataProviders()) {
+        xmlSuite.setShareThreadPoolForDataProviders(true);
+      }
+      if (m_configuration.useGlobalThreadPool()) {
+        xmlSuite.shouldUseGlobalThreadPool(true);
+      }
       createSuiteRunners(suiteRunnerMap, xmlSuite);
     }
 
@@ -1185,29 +1224,15 @@ public class TestNG {
     IThreadWorkerFactory<ISuite> factory =
         new SuiteWorkerFactory(
             suiteRunnerMap, 0 /* verbose hasn't been set yet */, getDefaultSuiteName());
-    ITestNGThreadPoolExecutor pooledExecutor =
-        this.getExecutorFactory()
-            .newSuiteExecutor(
-                "suites",
-                suiteGraph,
-                factory,
-                m_suiteThreadPoolSize,
-                m_suiteThreadPoolSize,
-                Integer.MAX_VALUE,
-                TimeUnit.MILLISECONDS,
-                new LinkedBlockingQueue<>(),
-                null);
-
-    Utils.log("TestNG", 2, "Starting executor for all suites");
-    // Run all suites in parallel
-    pooledExecutor.run();
-    try {
-      pooledExecutor.awaitTermination(Long.MAX_VALUE, TimeUnit.MILLISECONDS);
-      pooledExecutor.shutdownNow();
-    } catch (InterruptedException handled) {
-      Thread.currentThread().interrupt();
-      error("Error waiting for concurrent executors to finish " + handled.getMessage());
-    }
+    SuiteTaskExecutor taskExecutor =
+        new SuiteTaskExecutor(
+            this.m_configuration,
+            factory,
+            new LinkedBlockingQueue<>(),
+            suiteGraph,
+            m_suiteThreadPoolSize);
+    taskExecutor.execute();
+    taskExecutor.awaitCompletion();
 
     //
     // Generate the suites report
@@ -1280,10 +1305,6 @@ public class TestNG {
    * @param xmlSuite Xml Suite (and its children) for which {@code SuiteRunner}s are created
    */
   private void createSuiteRunners(SuiteRunnerMap suiteRunnerMap /* OUT */, XmlSuite xmlSuite) {
-    if (null != m_isJUnit && !m_isJUnit.equals(XmlSuite.DEFAULT_JUNIT)) {
-      xmlSuite.setJUnit(m_isJUnit);
-    }
-
     // If the skip flag was invoked on the command line, it
     // takes precedence
     if (null != m_skipFailedInvocationCounts) {
@@ -1324,9 +1345,11 @@ public class TestNG {
 
   /** Creates a suite runner and configures its initial state */
   private SuiteRunner createSuiteRunner(XmlSuite xmlSuite) {
-    DataProviderHolder holder = new DataProviderHolder();
+    DataProviderHolder holder = new DataProviderHolder(m_configuration);
     holder.addListeners(m_dataProviderListeners.values());
     holder.addInterceptors(m_dataProviderInterceptors.values());
+    TestListenersContainer container =
+        new TestListenersContainer(getTestListeners(), this.exitCodeListener);
     SuiteRunner result =
         new SuiteRunner(
             getConfiguration(),
@@ -1336,10 +1359,10 @@ public class TestNG {
             m_useDefaultListeners,
             m_methodInterceptors,
             m_invokedMethodListeners.values(),
-            m_testListeners.values(),
+            container,
             m_classListeners.values(),
             holder,
-            Systematiser.getComparator());
+            MethodSorting.basedOn());
 
     for (ISuiteListener isl : m_suiteListeners.values()) {
       result.addListener(isl);
@@ -1423,7 +1446,29 @@ public class TestNG {
    * @param cla The command line parameters
    */
   protected void configure(CommandLineArgs cla) {
+    Optional.ofNullable(cla.useGlobalThreadPool).ifPresent(this::shouldUseGlobalThreadPool);
+    Optional.ofNullable(cla.shareThreadPoolForDataProviders)
+        .ifPresent(this::shareThreadPoolForDataProviders);
+    Optional.ofNullable(cla.propagateDataProviderFailureAsTestFailure)
+        .ifPresent(value -> propagateDataProviderFailureAsTestFailure());
     setReportAllDataDrivenTestsAsSkipped(cla.includeAllDataDrivenTestsWhenSkipping);
+
+    Optional.ofNullable(cla.listenerFactory)
+        .map(ClassHelper::forName)
+        .filter(ITestNGListenerFactory.class::isAssignableFrom)
+        .map(it -> m_objectFactory.newInstance(it))
+        .map(it -> (ITestNGListenerFactory) it)
+        .ifPresent(this::setListenerFactory);
+
+    Optional.ofNullable(cla.generateResultsPerSuite).ifPresent(this::setGenerateResultsPerSuite);
+
+    Optional.ofNullable(cla.listenerComparator)
+        .map(ClassHelper::forName)
+        .filter(ListenerComparator.class::isAssignableFrom)
+        .map(it -> m_objectFactory.newInstance(it))
+        .map(it -> (ListenerComparator) it)
+        .ifPresent(this::setListenerComparator);
+
     if (cla.verbose != null) {
       setVerbose(cla.verbose);
     }
@@ -1434,9 +1479,13 @@ public class TestNG {
             m_objectFactory.newInstance((Class<IInjectorFactory>) clazz));
       }
     }
-    if (cla.threadPoolFactoryClass != null) {
-      setExecutorFactoryClass(cla.threadPoolFactoryClass);
-    }
+    Optional.ofNullable(cla.threadPoolFactoryClass)
+        .map(ClassHelper::forName)
+        .filter(IExecutorServiceFactory.class::isAssignableFrom)
+        .map(it -> m_objectFactory.newInstance(it))
+        .map(it -> (IExecutorServiceFactory) it)
+        .ifPresent(this::setExecutorServiceFactory);
+
     setOutputDirectory(cla.outputDirectory);
 
     String testClasses = cla.testClass;
@@ -1454,6 +1503,7 @@ public class TestNG {
 
     if (cla.testNames != null) {
       setTestNames(Arrays.asList(cla.testNames.split(",")));
+      setIgnoreMissedTestNames(cla.ignoreMissedTestNames);
     }
 
     // Note: can't use a Boolean field here because we are allowing a boolean
@@ -1466,8 +1516,6 @@ public class TestNG {
     setExcludedGroups(cla.excludedGroups);
     setTestJar(cla.testJar);
     setXmlPathInJar(cla.xmlPathInJar);
-    setJUnit(cla.junit);
-    setMixed(cla.mixed);
     setSkipFailedInvocationCounts(cla.skipFailedInvocationCounts);
     toggleFailureIfAllTestsWereSkipped(cla.failIfAllTestsSkipped);
     setListenersToSkipFromBeingWiredInViaServiceLoaders(cla.spiListenersToSkip.split(","));
@@ -1540,7 +1588,7 @@ public class TestNG {
       addReporter(reporterConfig);
     }
 
-    if (cla.commandLineMethods.size() > 0) {
+    if (!cla.commandLineMethods.isEmpty()) {
       m_commandLineMethods = cla.commandLineMethods;
     }
 
@@ -1553,6 +1601,10 @@ public class TestNG {
     alwaysRunListeners(cla.alwaysRunListeners);
   }
 
+  private void setIgnoreMissedTestNames(boolean ignoreMissedTestNames) {
+    m_ignoreMissedTestNames = ignoreMissedTestNames;
+  }
+
   public void setSuiteThreadPoolSize(Integer suiteThreadPoolSize) {
     m_suiteThreadPoolSize = suiteThreadPoolSize;
   }
@@ -1631,7 +1683,6 @@ public class TestNG {
     result.excludedGroups = (String) cmdLineArgs.get(CommandLineArgs.EXCLUDED_GROUPS);
     result.testJar = (String) cmdLineArgs.get(CommandLineArgs.TEST_JAR);
     result.xmlPathInJar = (String) cmdLineArgs.get(CommandLineArgs.XML_PATH_IN_JAR);
-    result.junit = (Boolean) cmdLineArgs.get(CommandLineArgs.JUNIT);
     result.mixed = (Boolean) cmdLineArgs.get(CommandLineArgs.MIXED);
     Object tmpValue = cmdLineArgs.get(CommandLineArgs.INCLUDE_ALL_DATA_DRIVEN_TESTS_WHEN_SKIPPING);
     if (tmpValue != null) {
@@ -1714,6 +1765,18 @@ public class TestNG {
       result.dependencyInjectorFactoryClass = dependencyInjectorFactoryClass;
     }
 
+    result.ignoreMissedTestNames =
+        Boolean.parseBoolean(
+            cmdLineArgs.getOrDefault(CommandLineArgs.IGNORE_MISSED_TEST_NAMES, false).toString());
+
+    result.generateResultsPerSuite =
+        Boolean.parseBoolean(
+            cmdLineArgs.getOrDefault(CommandLineArgs.GENERATE_RESULTS_PER_SUITE, false).toString());
+
+    Optional.ofNullable(cmdLineArgs.get(CommandLineArgs.LISTENER_COMPARATOR))
+        .map(Object::toString)
+        .ifPresent(it -> result.listenerComparator = it);
+
     configure(result);
   }
 
@@ -1753,23 +1816,6 @@ public class TestNG {
     return reporter;
   }
 
-  /**
-   * Specify if this run should be made in JUnit mode
-   *
-   * @param isJUnit - Specify if this run should be made in JUnit mode
-   */
-  public void setJUnit(Boolean isJUnit) {
-    m_isJUnit = isJUnit;
-  }
-
-  /** @param isMixed Specify if this run should be made in mixed mode */
-  public void setMixed(Boolean isMixed) {
-    if (isMixed == null) {
-      return;
-    }
-    m_isMixed = isMixed;
-  }
-
   /**
    * Double check that the command line parameters are valid.
    *
@@ -1798,13 +1844,6 @@ public class TestNG {
         && (testNgXml == null || testNgXml.isEmpty())) {
       throw new ParameterException("Groups option should be used with testclass option");
     }
-
-    Boolean junit = args.junit;
-    Boolean mixed = args.mixed;
-    if (junit && mixed) {
-      throw new ParameterException(
-          CommandLineArgs.MIXED + " can't be combined with " + CommandLineArgs.JUNIT);
-    }
   }
 
   /** @return true if at least one test failed. */
