@@ -1,5 +1,6 @@
 package org.testng;
 
+import static org.testng.ListenerComparator.sort;
 import static org.testng.internal.MethodHelper.fixMethodsWithClass;
 
 import java.lang.reflect.Method;
@@ -10,19 +11,21 @@ import java.util.Comparator;
 import java.util.Date;
 import java.util.List;
 import java.util.Map;
+import java.util.Objects;
 import java.util.Optional;
 import java.util.Set;
 import java.util.concurrent.BlockingQueue;
 import java.util.concurrent.LinkedBlockingQueue;
 import java.util.concurrent.PriorityBlockingQueue;
-import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicReference;
+import java.util.stream.Collectors;
 import java.util.stream.Stream;
-import javax.annotation.Nonnull;
 import org.testng.collections.Lists;
 import org.testng.collections.Maps;
 import org.testng.collections.Sets;
 import org.testng.internal.Attributes;
+import org.testng.internal.BaseTestMethod;
+import org.testng.internal.ClassBasedWrapper;
 import org.testng.internal.ClassInfoMap;
 import org.testng.internal.ConfigurationGroupMethods;
 import org.testng.internal.DefaultListenerFactory;
@@ -33,16 +36,18 @@ import org.testng.internal.IConfiguration;
 import org.testng.internal.IContainer;
 import org.testng.internal.ITestClassConfigInfo;
 import org.testng.internal.ITestResultNotifier;
+import org.testng.internal.ListenerOrderDeterminer;
 import org.testng.internal.MethodGroupsHelper;
 import org.testng.internal.MethodHelper;
+import org.testng.internal.MethodSorting;
 import org.testng.internal.ResultMap;
 import org.testng.internal.RunInfo;
 import org.testng.internal.RuntimeBehavior;
-import org.testng.internal.Systematiser;
 import org.testng.internal.TestListenerHelper;
 import org.testng.internal.TestMethodComparator;
 import org.testng.internal.TestMethodContainer;
 import org.testng.internal.TestNGClassFinder;
+import org.testng.internal.TestNGDeadLockException;
 import org.testng.internal.TestNGMethodFinder;
 import org.testng.internal.Utils;
 import org.testng.internal.XmlMethodSelector;
@@ -52,18 +57,14 @@ import org.testng.internal.invokers.ConfigMethodArguments;
 import org.testng.internal.invokers.ConfigMethodArguments.Builder;
 import org.testng.internal.invokers.IInvoker;
 import org.testng.internal.invokers.Invoker;
-import org.testng.internal.invokers.TestMethodWorker;
 import org.testng.internal.objects.IObjectDispenser;
-import org.testng.junit.IJUnitTestRunner;
-import org.testng.log4testng.Logger;
-import org.testng.thread.ITestNGThreadPoolExecutor;
 import org.testng.thread.IThreadWorkerFactory;
 import org.testng.thread.IWorker;
 import org.testng.util.Strings;
 import org.testng.util.TimeUtils;
 import org.testng.xml.XmlClass;
-import org.testng.xml.XmlInclude;
 import org.testng.xml.XmlPackage;
+import org.testng.xml.XmlSuite;
 import org.testng.xml.XmlTest;
 
 /** This class takes care of running one Test. */
@@ -74,7 +75,6 @@ public class TestRunner
         IConfigEavesdropper {
 
   private static final String DEFAULT_PROP_OUTPUT_DIR = "test-output";
-  private static final Logger LOGGER = Logger.getLogger(TestRunner.class);
 
   private final Comparator<ITestNGMethod> comparator;
   private ISuite m_suite;
@@ -96,10 +96,9 @@ public class TestRunner
 
   private final IConfigurationListener m_confListener = new ConfigurationListener();
 
-  private Collection<IInvokedMethodListener> m_invokedMethodListeners = Lists.newArrayList();
   private final Map<Class<? extends IClassListener>, IClassListener> m_classListeners =
       Maps.newLinkedHashMap();
-  private final DataProviderHolder holder = new DataProviderHolder();
+  private final DataProviderHolder holder;
 
   private Date m_startDate = new Date();
   private Date m_endDate = null;
@@ -115,6 +114,8 @@ public class TestRunner
   // The XML method selector (groups/methods included/excluded in XML)
   private final XmlMethodSelector m_xmlMethodSelector = new XmlMethodSelector();
 
+  private ITestListener exitCodeListener;
+
   //
   // These next fields contain all the configuration methods found on this class.
   // At initialization time, they just contain all the various @Configuration methods
@@ -170,9 +171,10 @@ public class TestRunner
       Collection<IInvokedMethodListener> invokedMethodListeners,
       List<IClassListener> classListeners,
       Comparator<ITestNGMethod> comparator,
-      DataProviderHolder otherHolder) {
+      DataProviderHolder otherHolder,
+      ISuiteRunnerListener suiteRunner) {
     this.comparator = comparator;
-    this.holder.merge(otherHolder);
+    this.holder = otherHolder;
     init(
         configuration,
         suite,
@@ -181,7 +183,8 @@ public class TestRunner
         finder,
         skipFailedInvocationCounts,
         invokedMethodListeners,
-        classListeners);
+        classListeners,
+        suiteRunner);
   }
 
   public TestRunner(
@@ -191,8 +194,10 @@ public class TestRunner
       boolean skipFailedInvocationCounts,
       Collection<IInvokedMethodListener> invokedMethodListeners,
       List<IClassListener> classListeners,
-      Comparator<ITestNGMethod> comparator) {
+      Comparator<ITestNGMethod> comparator,
+      ISuiteRunnerListener suiteRunner) {
     this.comparator = comparator;
+    this.holder = new DataProviderHolder(configuration);
     init(
         configuration,
         suite,
@@ -201,7 +206,8 @@ public class TestRunner
         suite.getAnnotationFinder(),
         skipFailedInvocationCounts,
         invokedMethodListeners,
-        classListeners);
+        classListeners,
+        suiteRunner);
   }
 
   /* /!\ This constructor is used by testng-remote, any changes related to it please contact with testng-team. */
@@ -211,8 +217,10 @@ public class TestRunner
       XmlTest test,
       boolean skipFailedInvocationCounts,
       Collection<IInvokedMethodListener> invokedMethodListeners,
-      List<IClassListener> classListeners) {
-    this.comparator = Systematiser.getComparator();
+      List<IClassListener> classListeners,
+      ISuiteRunnerListener suiteRunner) {
+    this.comparator = MethodSorting.basedOn();
+    this.holder = new DataProviderHolder(configuration);
     init(
         configuration,
         suite,
@@ -221,7 +229,8 @@ public class TestRunner
         suite.getAnnotationFinder(),
         skipFailedInvocationCounts,
         invokedMethodListeners,
-        classListeners);
+        classListeners,
+        suiteRunner);
   }
 
   private void init(
@@ -232,7 +241,8 @@ public class TestRunner
       IAnnotationFinder annotationFinder,
       boolean skipFailedInvocationCounts,
       Collection<IInvokedMethodListener> invokedMethodListeners,
-      List<IClassListener> classListeners) {
+      List<IClassListener> classListeners,
+      ISuiteRunnerListener suiteRunner) {
     m_configuration = configuration;
     m_xmlTest = test;
     m_suite = suite;
@@ -242,6 +252,13 @@ public class TestRunner
     m_injectorFactory = m_configuration.getInjectorFactory();
     m_objectFactory = suite.getObjectFactory();
     setVerbose(test.getVerbose());
+    if (suiteRunner == null) {
+      if (suite instanceof ISuiteRunnerListener) {
+        setExitCodeListener(((ISuiteRunnerListener) suite).getExitCodeListener());
+      }
+    } else {
+      setExitCodeListener(suiteRunner.getExitCodeListener());
+    }
 
     boolean preserveOrder = test.getPreserveOrder();
     IMethodInterceptor builtinInterceptor =
@@ -249,7 +266,7 @@ public class TestRunner
             ? new PreserveOrderMethodInterceptor()
             : new InstanceOrderingMethodInterceptor();
     m_methodInterceptors = new ArrayList<>();
-    // Add the built in interceptor as the first interceptor. That way we let our users determine
+    // Add the built-in interceptor as the first interceptor. That way we let our users determine
     // the final order
     // by plugging in their own custom interceptors as well.
     m_methodInterceptors.add(builtinInterceptor);
@@ -260,7 +277,6 @@ public class TestRunner
     }
 
     m_annotationFinder = annotationFinder;
-    m_invokedMethodListeners = invokedMethodListeners;
     m_classListeners.clear();
     for (IClassListener classListener : classListeners) {
       m_classListeners.put(classListener.getClass(), classListener);
@@ -274,7 +290,9 @@ public class TestRunner
             skipFailedInvocationCounts,
             invokedMethodListeners,
             classListeners,
-            holder);
+            holder,
+            m_confListener,
+            suiteRunner);
 
     if (test.getParallel() != null) {
       log("Running the tests in '" + test.getName() + "' with parallel mode:" + test.getParallel());
@@ -328,13 +346,9 @@ public class TestRunner
     initRunInfo(m_xmlTest);
 
     // Init methods and class map
-    // JUnit behavior is different and doesn't need this initialization step
-    if (!m_xmlTest.isJUnit()) {
-      initMethods();
-    }
+    initMethods();
 
     initListeners();
-    addConfigurationListener(m_confListener);
     for (IConfigurationListener cl : m_configuration.getConfigurationListeners()) {
       addConfigurationListener(cl);
     }
@@ -345,7 +359,7 @@ public class TestRunner
     // Find all the listener factories and collect all the listeners requested in a
     // @Listeners annotation.
     //
-    Set<Class<? extends ITestNGListener>> listenerClasses = Sets.newHashSet();
+    Set<Class<? extends ITestNGListener>> listenerClasses = Sets.newLinkedHashSet();
     Class<? extends ITestNGListenerFactory> listenerFactoryClass = null;
 
     for (IClass cls : getTestClasses()) {
@@ -369,16 +383,12 @@ public class TestRunner
     //
 
     ITestNGListenerFactory factory =
-        TestListenerHelper.createListenerFactory(
-            m_objectFactory, m_testClassFinder, listenerFactoryClass, this);
+        Optional.ofNullable(m_configuration.getListenerFactory())
+            .orElse(new DefaultListenerFactory(m_objectFactory, this));
 
     // Instantiate all the listeners
     for (Class<? extends ITestNGListener> c : listenerClasses) {
-      if (IClassListener.class.isAssignableFrom(c) && m_classListeners.containsKey(c)) {
-        continue;
-      }
       ITestNGListener listener = factory.createListener(c);
-
       addListener(listener);
     }
   }
@@ -478,15 +488,13 @@ public class TestRunner
     // Walk through all the TestClasses, store their method
     // and initialize them with the correct ITestClass
     //
+
     for (ITestClass tc : m_classMap.values()) {
       fixMethodsWithClass(tc.getTestMethods(), tc, testMethods);
-      fixMethodsWithClass(
-          ((ITestClassConfigInfo) tc).getAllBeforeClassMethods().toArray(new ITestNGMethod[0]),
-          tc,
-          beforeClassMethods);
+      fixMethodsWithClass(beforeClassConfigMethods(tc), tc, beforeClassMethods);
       fixMethodsWithClass(tc.getBeforeTestMethods(), tc, null);
       fixMethodsWithClass(tc.getAfterTestMethods(), tc, null);
-      fixMethodsWithClass(tc.getAfterClassMethods(), tc, afterClassMethods);
+      fixMethodsWithClass(afterClassConfigMethods(tc), tc, afterClassMethods);
       fixMethodsWithClass(tc.getBeforeSuiteMethods(), tc, beforeSuiteMethods);
       fixMethodsWithClass(tc.getAfterSuiteMethods(), tc, afterSuiteMethods);
       fixMethodsWithClass(tc.getBeforeTestConfigurationMethods(), tc, beforeXmlTestMethods);
@@ -550,6 +558,14 @@ public class TestRunner
             comparator);
   }
 
+  private static ITestNGMethod[] beforeClassConfigMethods(ITestClass tc) {
+    return ITestClassConfigInfo.allBeforeClassMethods(tc).toArray(ITestNGMethod[]::new);
+  }
+
+  private static ITestNGMethod[] afterClassConfigMethods(ITestClass tc) {
+    return ITestClassConfigInfo.allAfterClassMethods(tc).toArray(ITestNGMethod[]::new);
+  }
+
   private ITestNGMethod[] computeAndGetAllTestMethods() {
     List<ITestNGMethod> testMethods = Lists.newArrayList();
     for (ITestClass tc : m_classMap.values()) {
@@ -597,11 +613,7 @@ public class TestRunner
 
     try {
       XmlTest test = getTest();
-      if (test.isJUnit()) {
-        privateRunJUnit();
-      } else {
-        privateRun(test);
-      }
+      privateRun(test);
     } finally {
       afterRun();
       forgetHeavyReferencesIfNeeded();
@@ -646,71 +658,6 @@ public class TestRunner
     }
   }
 
-  private ITestNGMethod[] m_allJunitTestMethods = new ITestNGMethod[] {};
-
-  private void privateRunJUnit() {
-    final ClassInfoMap cim = new ClassInfoMap(m_testClassesFromXml, false);
-    final Set<Class<?>> classes = cim.getClasses();
-    final List<ITestNGMethod> runMethods = Lists.newArrayList();
-    List<IWorker<ITestNGMethod>> workers = Lists.newArrayList();
-    // FIXME: directly referencing JUnitTestRunner which uses JUnit classes
-    // may result in an class resolution exception under different JVMs
-    // The resolution process is not specified in the JVM spec with a specific implementation,
-    // so it can be eager => failure
-    workers.add(
-        new IWorker<ITestNGMethod>() {
-          /** @see TestMethodWorker#getTimeOut() */
-          @Override
-          public long getTimeOut() {
-            return 0;
-          }
-
-          /** @see java.lang.Runnable#run() */
-          @Override
-          public void run() {
-            for (Class<?> tc : classes) {
-              List<XmlInclude> includedMethods = cim.getXmlClass(tc).getIncludedMethods();
-              List<String> methods = Lists.newArrayList();
-              for (XmlInclude inc : includedMethods) {
-                methods.add(inc.getName());
-              }
-              IJUnitTestRunner tr =
-                  IJUnitTestRunner.createTestRunner(m_objectFactory, TestRunner.this);
-              tr.setInvokedMethodListeners(m_invokedMethodListeners);
-              try {
-                tr.run(tc, methods.toArray(new String[0]));
-              } catch (Exception ex) {
-                LOGGER.error(ex.getMessage(), ex);
-              } finally {
-                runMethods.addAll(tr.getTestMethods());
-              }
-            }
-          }
-
-          @Override
-          public List<ITestNGMethod> getTasks() {
-            throw new TestNGException("JUnit not supported");
-          }
-
-          @Override
-          public int getPriority() {
-            if (m_allJunitTestMethods.length == 1) {
-              return m_allJunitTestMethods[0].getPriority();
-            } else {
-              return 0;
-            }
-          }
-
-          @Override
-          public int compareTo(@Nonnull IWorker<ITestNGMethod> other) {
-            return getPriority() - other.getPriority();
-          }
-        });
-
-    runJUnitWorkers(workers);
-    m_allJunitTestMethods = runMethods.toArray(new ITestNGMethod[0]);
-  }
-
   private static Comparator<ITestNGMethod> newComparator(boolean needPrioritySort) {
     return needPrioritySort ? new TestMethodComparator() : null;
   }
@@ -732,8 +679,6 @@ public class TestRunner
   private void privateRun(XmlTest xmlTest) {
     boolean parallel = xmlTest.getParallel().isParallel();
 
-    // parallel
-    int threadCount = parallel ? xmlTest.getThreadCount() : 1;
     // Make sure we create a graph based on the intercepted methods, otherwise an interceptor
     // removing methods would cause the graph never to terminate (because it would expect
     // termination from methods that never get invoked).
@@ -748,7 +693,20 @@ public class TestRunner
         });
     IDynamicGraph<ITestNGMethod> graph = reference.get();
 
-    graph.setVisualisers(this.visualisers);
+    for (ITestNGMethod each : interceptedOrder) {
+      if (each instanceof BaseTestMethod) {
+        // We don't want our users to change this vital info. That is why the setter is NOT
+        // being exposed via the interface, and so we resort to an "instanceof" check.
+        Set<ITestNGMethod> downstream = Sets.newHashSet(graph.getDependenciesFor(each));
+        ((BaseTestMethod) each).setDownstreamDependencies(downstream);
+        Set<ITestNGMethod> upstream = Sets.newHashSet(graph.getUpstreamDependenciesFor(each));
+        ((BaseTestMethod) each).setUpstreamDependencies(upstream);
+      }
+    }
+
+    Collection<IExecutionVisualiser> original =
+        sort(this.visualisers, m_configuration.getListenerComparator());
+    graph.setVisualisers(Sets.newLinkedHashSet(original));
     // In some cases, additional sorting is needed to make sure tests run in the appropriate order.
     // If the user specified a method interceptor, or if we have any methods that have a non-default
     // priority on them, we need to sort.
@@ -758,36 +716,11 @@ public class TestRunner
       if (graph.getNodeCount() <= 0) {
         return;
       }
-      ITestNGThreadPoolExecutor executor =
-          this.m_configuration
-              .getExecutorFactory()
-              .newTestMethodExecutor(
-                  "test=" + xmlTest.getName(),
-                  graph,
-                  this,
-                  threadCount,
-                  threadCount,
-                  0,
-                  TimeUnit.MILLISECONDS,
-                  newQueue(needPrioritySort),
-                  methodComparator);
-      executor.run();
-      try {
-        long timeOut = m_xmlTest.getTimeOut(XmlTest.DEFAULT_TIMEOUT_MS);
-        Utils.log(
-            "TestRunner",
-            2,
-            "Starting executor for test "
-                + m_xmlTest.getName()
-                + " with time out:"
-                + timeOut
-                + " milliseconds.");
-        executor.awaitTermination(timeOut, TimeUnit.MILLISECONDS);
-        executor.shutdownNow();
-      } catch (InterruptedException handled) {
-        LOGGER.error(handled.getMessage(), handled);
-        Thread.currentThread().interrupt();
-      }
+      TestTaskExecutor taskExecutor =
+          new TestTaskExecutor(
+              m_configuration, xmlTest, this, newQueue(needPrioritySort), graph, methodComparator);
+      taskExecutor.execute();
+      taskExecutor.awaitCompletion();
       return;
     }
     List<ITestNGMethod> freeNodes = graph.getFreeNodes();
@@ -817,7 +750,9 @@ public class TestRunner
     List<IMethodInstance> methodInstances =
         MethodHelper.methodsToMethodInstances(Arrays.asList(methods));
 
-    for (IMethodInterceptor m_methodInterceptor : m_methodInterceptors) {
+    List<IMethodInterceptor> original =
+        sort(m_methodInterceptors, m_configuration.getListenerComparator());
+    for (IMethodInterceptor m_methodInterceptor : original) {
       methodInstances = m_methodInterceptor.intercept(methodInstances, this);
     }
 
@@ -874,17 +809,34 @@ public class TestRunner
             .testContext(this)
             .listeners(this.m_classListeners.values())
             .build();
-    return AbstractParallelWorker.newWorker(
-            m_xmlTest.getParallel(), m_xmlTest.getGroupByInstances())
-        .createWorkers(args);
-  }
-
-  //
-  // Invoke the workers
-  //
-  private void runJUnitWorkers(List<? extends IWorker<ITestNGMethod>> workers) {
-    // Sequential run
-    workers.forEach(Runnable::run);
+    List<IWorker<ITestNGMethod>> result =
+        AbstractParallelWorker.newWorker(m_xmlTest.getParallel(), m_xmlTest.getGroupByInstances())
+            .createWorkers(args);
+    long dataDrivenTestCount =
+        result.stream()
+            .flatMap(it -> it.getTasks().stream())
+            .filter(ITestNGMethod::isDataDriven)
+            .count();
+    int threads = getCurrentXmlTest().getThreadCount();
+    XmlSuite.ParallelMode parallelMode = getCurrentXmlTest().getParallel();
+    XmlSuite suite = getSuite().getXmlSuite();
+    if (suite.useGlobalThreadPool()
+        && parallelMode.isParallel()
+        && dataDrivenTestCount >= threads) {
+      String msg =
+          "[Deadlock condition detected] "
+              + "Cannot run "
+              + dataDrivenTestCount
+              + " data driven tests on just "
+              + threads
+              + " threads when "
+              + "using common thread pool. "
+              + "Please increase the number of threads to at-least "
+              + (dataDrivenTestCount + 1)
+              + ".";
+      throw new TestNGDeadLockException(msg);
+    }
+    return result;
   }
 
   private void afterRun() {
@@ -934,15 +886,20 @@ public class TestRunner
    */
   private void fireEvent(boolean isStart) {
     if (isStart) {
-      for (ITestListener itl : m_testListeners) {
+      for (ITestListener itl :
+          ListenerOrderDeterminer.order(m_testListeners, m_configuration.getListenerComparator())) {
         itl.onStart(this);
       }
+      this.exitCodeListener.onStart(this);
 
     } else {
-      List<ITestListener> testListenersReversed = Lists.newReversedArrayList(m_testListeners);
+      List<ITestListener> testListenersReversed =
+          ListenerOrderDeterminer.reversedOrder(
+              m_testListeners, m_configuration.getListenerComparator());
       for (ITestListener itl : testListenersReversed) {
         itl.onFinish(this);
       }
+      this.exitCodeListener.onFinish(this);
     }
     if (!isStart) {
       MethodHelper.clear(methods(this.getPassedConfigurations()));
@@ -1025,10 +982,6 @@ public class TestRunner
 
   @Override
   public ITestNGMethod[] getAllTestMethods() {
-    if (getTest().isJUnit()) {
-      // This is true only when we are running JUnit mode
-      return m_allJunitTestMethods;
-    }
     return testMethodsContainer.getItems();
   }
 
@@ -1119,20 +1072,11 @@ public class TestRunner
 
   @Override
   public List<IConfigurationListener> getConfigurationListeners() {
-    List<IConfigurationListener> listeners = Lists.newArrayList(m_configurationListeners);
-    for (IConfigurationListener each : this.m_configuration.getConfigurationListeners()) {
-      boolean duplicate = false;
-      for (IConfigurationListener listener : listeners) {
-        if (each.getClass().equals(listener.getClass())) {
-          duplicate = true;
-          break;
-        }
-      }
-      if (!duplicate) {
-        listeners.add(each);
-      }
-    }
-    return Lists.newArrayList(listeners);
+    return m_configurationListeners.stream()
+        .map(ClassBasedWrapper::wrap)
+        .distinct()
+        .map(ClassBasedWrapper::unWrap)
+        .collect(Collectors.toUnmodifiableList());
   }
 
   private void logFailedTest(ITestResult tr, boolean withinSuccessPercentage) {
@@ -1159,18 +1103,15 @@ public class TestRunner
   // TODO: This method needs to be removed and we need to be leveraging addListener().
   // Investigate and fix this.
   void addTestListener(ITestListener listener) {
-    Optional<ITestListener> found =
+    boolean found =
         m_testListeners.stream()
-            .filter(iTestListener -> iTestListener.getClass().equals(listener.getClass()))
-            .findAny();
-    if (found.isPresent()) {
-      return;
+            .anyMatch(iTestListener -> iTestListener.getClass().equals(listener.getClass()));
+    if (!found) {
+      m_testListeners.add(listener);
     }
-    m_testListeners.add(listener);
   }
 
   public void addListener(ITestNGListener listener) {
-    // TODO a listener may be added many times if it implements many interfaces
     if (listener instanceof IMethodInterceptor) {
       m_methodInterceptors.add((IMethodInterceptor) listener);
     }
@@ -1180,9 +1121,7 @@ public class TestRunner
     }
     if (listener instanceof IClassListener) {
       IClassListener classListener = (IClassListener) listener;
-      if (!m_classListeners.containsKey(classListener.getClass())) {
-        m_classListeners.put(classListener.getClass(), classListener);
-      }
+      m_classListeners.putIfAbsent(classListener.getClass(), classListener);
     }
     if (listener instanceof IConfigurationListener) {
       addConfigurationListener((IConfigurationListener) listener);
@@ -1216,7 +1155,20 @@ public class TestRunner
   }
 
   void addConfigurationListener(IConfigurationListener icl) {
-    m_configurationListeners.add(icl);
+    boolean alreadyAdded =
+        m_configurationListeners.stream().anyMatch(each -> each.getClass().equals(icl.getClass()));
+    if (!alreadyAdded) {
+      m_configurationListeners.add(icl);
+    }
+  }
+
+  private void setExitCodeListener(ITestListener exitCodeListener) {
+    this.exitCodeListener = exitCodeListener;
+  }
+
+  @Override
+  public ITestListener getExitCodeListener() {
+    return Objects.requireNonNull(exitCodeListener, "ExitCodeListener cannot be null.");
   }
 
   private void dumpInvokedMethods() {
