@@ -3,13 +3,14 @@ package org.testng.internal;
 import java.lang.reflect.Method;
 import java.util.List;
 import org.testng.IDataProviderMethod;
+import org.testng.IRetryDataProvider;
 import org.testng.annotations.IDataProviderAnnotation;
 
 /** Represents an @{@link org.testng.annotations.DataProvider} annotated method. */
 class DataProviderMethod implements IDataProviderMethod {
 
-  private final Object instance;
-  private final Method method;
+  protected Object instance;
+  protected Method method;
   private final IDataProviderAnnotation annotation;
 
   DataProviderMethod(Object instance, Method method, IDataProviderAnnotation annotation) {
@@ -42,4 +43,19 @@ class DataProviderMethod implements IDataProviderMethod {
   public List<Integer> getIndices() {
     return annotation.getIndices();
   }
+
+  @Override
+  public boolean propagateFailureAsTestFailure() {
+    return annotation.isPropagateFailureAsTestFailure();
+  }
+
+  @Override
+  public Class<? extends IRetryDataProvider> retryUsing() {
+    return annotation.retryUsing();
+  }
+
+  @Override
+  public boolean cacheDataForTestRetries() {
+    return annotation.isCacheDataForTestRetries();
+  }
 }
