@@ -7,24 +7,39 @@ import static org.testng.internal.invokers.Invoker.SAME_CLASS;
 import java.lang.reflect.InvocationTargetException;
 import java.util.Arrays;
 import java.util.Collection;
-import java.util.List;
+import java.util.HashSet;
 import java.util.Map;
 import java.util.Set;
 import java.util.concurrent.ConcurrentHashMap;
+import org.testng.ConfigurationNotInvokedException;
 import org.testng.IClass;
 import org.testng.IConfigurable;
+import org.testng.IConfigurationListener;
 import org.testng.IInvokedMethodListener;
+import org.testng.ISuiteRunnerListener;
+import org.testng.ITestClass;
 import org.testng.ITestContext;
 import org.testng.ITestNGMethod;
 import org.testng.ITestResult;
+import org.testng.ListenerComparator;
 import org.testng.Reporter;
 import org.testng.SuiteRunState;
 import org.testng.TestNGException;
 import org.testng.annotations.IConfigurationAnnotation;
-import org.testng.collections.Lists;
 import org.testng.collections.Maps;
 import org.testng.collections.Sets;
-import org.testng.internal.*;
+import org.testng.internal.AutoCloseableLock;
+import org.testng.internal.ClassHelper;
+import org.testng.internal.ConfigurationMethod;
+import org.testng.internal.ConstructorOrMethod;
+import org.testng.internal.IConfiguration;
+import org.testng.internal.ITestResultNotifier;
+import org.testng.internal.MethodHelper;
+import org.testng.internal.Parameters;
+import org.testng.internal.RuntimeBehavior;
+import org.testng.internal.TestListenerHelper;
+import org.testng.internal.TestResult;
+import org.testng.internal.Utils;
 import org.testng.internal.annotations.AnnotationHelper;
 import org.testng.internal.invokers.ConfigMethodArguments.Builder;
 import org.testng.internal.thread.ThreadUtil;
@@ -38,22 +53,35 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
       Maps.newConcurrentMap();
 
   private final boolean m_continueOnFailedConfiguration;
+  private boolean m_hasTestTagLevelFailures = false;
+  private boolean m_hasClassLevelFailures = false;
+  private boolean m_hasTestMethodLevelFailures = false;
 
   private final Set<ITestNGMethod> m_executedConfigMethods = ConcurrentHashMap.newKeySet();
 
   /** Group failures must be synced as the Invoker is accessed concurrently */
   private final Map<String, Boolean> m_beforegroupsFailures = Maps.newConcurrentMap();
 
+  private final IConfigurationListener internalConfigurationListener;
+
   public ConfigInvoker(
       ITestResultNotifier notifier,
       Collection<IInvokedMethodListener> invokedMethodListeners,
       ITestContext testContext,
       SuiteRunState suiteState,
-      IConfiguration configuration) {
-    super(notifier, invokedMethodListeners, testContext, suiteState, configuration);
+      IConfiguration configuration,
+      IConfigurationListener internalConfigurationListener,
+      ISuiteRunnerListener suiteRunner) {
+    super(notifier, invokedMethodListeners, testContext, suiteState, configuration, suiteRunner);
     this.m_continueOnFailedConfiguration =
         testContext.getSuite().getXmlSuite().getConfigFailurePolicy()
             == XmlSuite.FailurePolicy.CONTINUE;
+    this.internalConfigurationListener = internalConfigurationListener;
+  }
+
+  @Override
+  public IConfiguration getConfiguration() {
+    return this.m_configuration;
   }
 
   /**
@@ -62,34 +90,49 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
    */
   public boolean hasConfigurationFailureFor(
       ITestNGMethod testNGMethod, String[] groups, IClass testClass, Object instance) {
+    return hasConfigurationFailureFor(null, testNGMethod, groups, testClass, instance);
+  }
+
+  @Override
+  public boolean hasConfigurationFailureFor(
+      ITestNGMethod configMethod,
+      ITestNGMethod testNGMethod,
+      String[] groups,
+      IClass testClass,
+      Object instance) {
     boolean result = false;
 
     Class<?> cls = testClass.getRealClass();
 
     if (m_suiteState.isFailed()) {
-      result = true;
-    } else {
-      boolean hasConfigurationFailures = classConfigurationFailed(cls, instance);
-      if (hasConfigurationFailures) {
-        if (!m_continueOnFailedConfiguration) {
-          result = true;
-        } else {
-          Set<Object> set = getInvocationResults(testClass);
-          result = set.contains(instance);
-        }
-        return result;
+      // if there were suite level failures, then short circuit here itself and report them.
+      return true;
+    }
+
+    boolean annotationFound = canIgnoreConfigFailure(testClass, configMethod);
+    boolean hasConfigurationFailures = classConfigurationFailed(cls, instance);
+    if (hasConfigurationFailures) {
+      if (annotationFound) {
+        // We were told to ignore failures via the annotation.
+        return false;
       }
-      // if method is BeforeClass, currentTestMethod will be null
-      if (m_continueOnFailedConfiguration && hasConfigFailure(testNGMethod)) {
-        Object key = TestNgMethodUtils.getMethodInvocationToken(testNGMethod, instance);
-        result = m_methodInvocationResults.get(testNGMethod).contains(key);
-      } else if (!m_continueOnFailedConfiguration) {
-        for (Class<?> clazz : m_classInvocationResults.keySet()) {
-          if (clazz.isAssignableFrom(cls)
-              && m_classInvocationResults.get(clazz).contains(instance)) {
-            result = true;
-            break;
-          }
+      if (m_continueOnFailedConfiguration) {
+        Set<Object> set = getInvocationResults(testClass);
+        result = set.contains(instance);
+      } else {
+        result = true;
+      }
+      return result;
+    }
+    // if method is BeforeClass, currentTestMethod will be null
+    if ((m_continueOnFailedConfiguration || annotationFound) && hasConfigFailure(testNGMethod)) {
+      Object key = TestNgMethodUtils.getMethodInvocationToken(testNGMethod, instance);
+      result = m_methodInvocationResults.get(testNGMethod).contains(key);
+    } else if (!(m_continueOnFailedConfiguration || annotationFound)) {
+      for (Class<?> clazz : m_classInvocationResults.keySet()) {
+        if (clazz.isAssignableFrom(cls) && m_classInvocationResults.get(clazz).contains(instance)) {
+          result = true;
+          break;
         }
       }
     }
@@ -110,18 +153,14 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
    * @param arguments - A {@link GroupConfigMethodArguments} object.
    */
   public void invokeBeforeGroupsConfigurations(GroupConfigMethodArguments arguments) {
-    List<ITestNGMethod> filteredMethods = Lists.newArrayList();
     String[] groups = arguments.getTestMethod().getGroups();
 
-    for (String group : groups) {
-      List<ITestNGMethod> methods =
-          arguments.getGroupMethods().getBeforeGroupMethodsForGroup(group);
-      if (methods != null) {
-        filteredMethods.addAll(methods);
-      }
-    }
+    ITestNGMethod[] beforeMethodsArray =
+        arguments
+            .getGroupMethods()
+            .getBeforeGroupMethodsForGroup(groups)
+            .toArray(new ITestNGMethod[0]);
 
-    ITestNGMethod[] beforeMethodsArray = filteredMethods.toArray(new ITestNGMethod[0]);
     //
     // Invoke the right groups methods
     //
@@ -130,20 +169,19 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
           Arrays.stream(beforeMethodsArray)
               .filter(ConfigInvoker::isGroupLevelConfigurationMethod)
               .toArray(ITestNGMethod[]::new);
-      if (filteredConfigurations.length == 0) {
-        return;
+      if (filteredConfigurations.length != 0) {
+        // don't pass the IClass or the instance as the method may be external
+        // the invocation must be similar to @BeforeTest/@BeforeSuite
+        ConfigMethodArguments configMethodArguments =
+            new Builder()
+                .usingConfigMethodsAs(filteredConfigurations)
+                .forSuite(arguments.getSuite())
+                .usingParameters(arguments.getParameters())
+                .usingInstance(arguments.getInstance())
+                .forTestMethod(arguments.getTestMethod())
+                .build();
+        invokeConfigurations(configMethodArguments);
       }
-      // don't pass the IClass or the instance as the method may be external
-      // the invocation must be similar to @BeforeTest/@BeforeSuite
-      ConfigMethodArguments configMethodArguments =
-          new Builder()
-              .usingConfigMethodsAs(filteredConfigurations)
-              .forSuite(arguments.getSuite())
-              .usingParameters(arguments.getParameters())
-              .usingInstance(arguments.getInstance())
-              .forTestMethod(arguments.getTestMethod())
-              .build();
-      invokeConfigurations(configMethodArguments);
     }
 
     //
@@ -164,59 +202,30 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
       return;
     }
 
-    // See if the currentMethod is the last method in any of the groups
-    // it belongs to
-    Map<String, String> filteredGroups = Maps.newHashMap();
-    String[] groups = arguments.getTestMethod().getGroups();
-    for (String group : groups) {
-      if (arguments.getGroupMethods().isLastMethodForGroup(group, arguments.getTestMethod())) {
-        filteredGroups.put(group, group);
-      }
-    }
-
-    if (filteredGroups.isEmpty()) {
-      return;
-    }
-
-    // The list of afterMethods to run
-    Map<ITestNGMethod, ITestNGMethod> afterMethods = Maps.newHashMap();
-
-    // Now filteredGroups contains all the groups for which we need to run the afterGroups
-    // method.  Find all the methods that correspond to these groups and invoke them.
-    for (String g : filteredGroups.values()) {
-      List<ITestNGMethod> methods = arguments.getGroupMethods().getAfterGroupMethodsForGroup(g);
-      // Note:  should put them in a map if we want to make sure the same afterGroups
-      // doesn't get run twice
-      if (methods != null) {
-        for (ITestNGMethod m : methods) {
-          afterMethods.put(m, m);
-        }
-      }
-    }
-
     // Got our afterMethods, invoke them
+    Set<String> filteredGroups = new HashSet<>();
     ITestNGMethod[] filteredConfigurations =
-        afterMethods.keySet().stream()
+        arguments.getGroupMethods().getAfterGroupMethods(arguments.getTestMethod()).stream()
+            .peek(t -> filteredGroups.addAll(Arrays.asList(t.getGroups())))
             .filter(ConfigInvoker::isGroupLevelConfigurationMethod)
             .toArray(ITestNGMethod[]::new);
-    if (filteredConfigurations.length == 0) {
-      return;
-    }
-    // don't pass the IClass or the instance as the method may be external
-    // the invocation must be similar to @BeforeTest/@BeforeSuite
-    ConfigMethodArguments configMethodArguments =
-        new Builder()
-            .usingConfigMethodsAs(filteredConfigurations)
-            .forSuite(arguments.getSuite())
-            .usingParameters(arguments.getParameters())
-            .usingInstance(arguments.getInstance())
-            .forTestMethod(arguments.getTestMethod())
-            .build();
+    if (filteredConfigurations.length != 0) {
+      // don't pass the IClass or the instance as the method may be external
+      // the invocation must be similar to @BeforeTest/@BeforeSuite
+      ConfigMethodArguments configMethodArguments =
+          new Builder()
+              .usingConfigMethodsAs(filteredConfigurations)
+              .forSuite(arguments.getSuite())
+              .usingParameters(arguments.getParameters())
+              .usingInstance(arguments.getInstance())
+              .forTestMethod(arguments.getTestMethod())
+              .build();
 
-    invokeConfigurations(configMethodArguments);
+      invokeConfigurations(configMethodArguments);
+    }
 
     // Remove the groups so they don't get run again
-    arguments.getGroupMethods().removeAfterGroups(filteredGroups.keySet());
+    arguments.getGroupMethods().removeAfterGroups(filteredGroups);
   }
 
   public void invokeConfigurations(ConfigMethodArguments arguments) {
@@ -227,7 +236,7 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
 
     ITestNGMethod[] methods =
         TestNgMethodUtils.filterMethods(
-            arguments.getTestClass(), arguments.getConfigMethods(), SAME_CLASS);
+            null, arguments.getTestClass(), arguments.getConfigMethods(), SAME_CLASS);
     Object[] parameters = new Object[] {};
 
     for (ITestNGMethod tm : methods) {
@@ -269,6 +278,7 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
           continue;
         }
         if (hasConfigurationFailureFor(
+                tm,
                 arguments.getTestMethod(),
                 tm.getGroups(),
                 arguments.getTestClass(),
@@ -276,9 +286,12 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
             && !alwaysRun) {
           log(3, "Skipping " + Utils.detailedMethodName(tm, true));
           InvokedMethod invokedMethod = new InvokedMethod(System.currentTimeMillis(), testResult);
+          // Set test result as 'SKIP' in 'beforeConfiguration' & 'beforeInvocation' if
+          // config method is skip.
+          testResult.setStatus(ITestResult.SKIP);
+          runConfigurationListeners(testResult, arguments.getTestMethod(), true /* before */);
           runInvokedMethodListeners(BEFORE_INVOCATION, invokedMethod, testResult);
           testResult.setEndMillis(testResult.getStartMillis());
-          testResult.setStatus(ITestResult.SKIP);
           runInvokedMethodListeners(AFTER_INVOCATION, invokedMethod, testResult);
 
           handleConfigurationSkip(
@@ -311,7 +324,8 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
         runConfigurationListeners(testResult, arguments.getTestMethod(), true /* before */);
 
         Object newInstance = computeInstance(arguments.getInstance(), inst, tm);
-        if (isConfigMethodEligibleForScrutiny(tm)) {
+        boolean isFirstTimeOnlyConfigMethod = isConfigMethodEligibleForScrutiny(tm);
+        if (isFirstTimeOnlyConfigMethod) {
           if (m_executedConfigMethods.add(arguments.getTestMethod())) {
             invokeConfigurationMethod(newInstance, tm, parameters, testResult);
           }
@@ -319,7 +333,9 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
           invokeConfigurationMethod(newInstance, tm, parameters, testResult);
         }
         copyAttributesFromNativelyInjectedTestResult(parameters, arguments.getTestMethodResult());
-        runConfigurationListeners(testResult, arguments.getTestMethod(), false /* after */);
+        if (!isFirstTimeOnlyConfigMethod) {
+          runConfigurationListeners(testResult, arguments.getTestMethod(), false /* after */);
+        }
         if (testResult.getStatus() == ITestResult.SKIP) {
           Throwable t = testResult.getThrowable();
           if (t != null) {
@@ -374,15 +390,28 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
         testResult.setStatus(ITestResult.SUCCESS);
         return;
       }
-      if (configurableInstance != null) {
-        MethodInvocationHelper.invokeConfigurable(
-            targetInstance, params, configurableInstance, method.getMethod(), testResult);
+      boolean willfullyIgnored = false;
+      boolean usesConfigurableInstance = configurableInstance != null;
+      if (usesConfigurableInstance) {
+        willfullyIgnored =
+            !MethodInvocationHelper.invokeConfigurable(
+                targetInstance, params, configurableInstance, method.getMethod(), testResult);
       } else {
         MethodInvocationHelper.invokeMethodConsideringTimeout(
-            tm, method, targetInstance, params, testResult);
+            tm, method, targetInstance, params, testResult, m_configuration);
+      }
+      boolean testStatusRemainedUnchanged = testResult.isNotRunning();
+      boolean throwException = !RuntimeBehavior.ignoreCallbackInvocationSkips();
+      if (throwException
+          && usesConfigurableInstance
+          && willfullyIgnored
+          && testStatusRemainedUnchanged) {
+        throw new ConfigurationNotInvokedException(tm);
       }
       testResult.setStatus(ITestResult.SUCCESS);
-    } catch (InvocationTargetException | IllegalAccessException ex) {
+    } catch (ConfigurationNotInvokedException
+        | InvocationTargetException
+        | IllegalAccessException ex) {
       throwConfigurationFailure(testResult, ex);
       testResult.setStatus(ITestResult.FAILURE);
       throw ex;
@@ -411,12 +440,21 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
   }
 
   private void runConfigurationListeners(ITestResult tr, ITestNGMethod tm, boolean before) {
+    ListenerComparator comparator = m_configuration.getListenerComparator();
     if (before) {
       TestListenerHelper.runPreConfigurationListeners(
-          tr, tm, m_notifier.getConfigurationListeners());
+          tr,
+          tm,
+          m_notifier.getConfigurationListeners(),
+          internalConfigurationListener,
+          comparator);
     } else {
       TestListenerHelper.runPostConfigurationListeners(
-          tr, tm, m_notifier.getConfigurationListeners());
+          tr,
+          tm,
+          m_notifier.getConfigurationListeners(),
+          internalConfigurationListener,
+          comparator);
     }
   }
 
@@ -492,7 +530,7 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
             classSetEntry -> {
               Set<Object> obj = classSetEntry.getValue();
               Class<?> c = classSetEntry.getKey();
-              boolean containsBeforeTestOrBeforeSuiteFailure = obj.contains(null);
+              boolean containsBeforeTestOrBeforeSuiteFailure = obj.contains(NULL_OBJECT);
               return c == cls
                   || c.isAssignableFrom(cls)
                       && (obj.contains(instance) || containsBeforeTestOrBeforeSuiteFailure);
@@ -519,11 +557,14 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
     instances.add(TestNgMethodUtils.getMethodInvocationToken(method, instance));
   }
 
+  private final AutoCloseableLock internalLock = new AutoCloseableLock();
+
   private void setClassInvocationFailure(Class<?> clazz, Object instance) {
-    synchronized (m_classInvocationResults) {
+    try (AutoCloseableLock ignore = internalLock.lock()) {
       Set<Object> instances =
           m_classInvocationResults.computeIfAbsent(clazz, k -> Sets.newHashSet());
-      instances.add(instance);
+      Object objectToAdd = instance == null ? NULL_OBJECT : instance;
+      instances.add(objectToAdd);
     }
   }
 
@@ -544,10 +585,13 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
     if (annotation.getBeforeTestClass() || annotation.getAfterTestClass()) {
       // tm is the configuration method, and currentTestMethod is null for BeforeClass
       // methods, so we need testClass
-      if (m_continueOnFailedConfiguration) {
-        setClassInvocationFailure(testClass.getRealClass(), instance);
-      } else {
-        setClassInvocationFailure(tm.getRealClass(), instance);
+      Class<?> clazzToUse = tm.getRealClass();
+      if (m_continueOnFailedConfiguration || annotation.isIgnoreFailure()) {
+        clazzToUse = testClass.getRealClass();
+      }
+      setClassInvocationFailure(clazzToUse, instance);
+      if (annotation.getBeforeTestClass()) {
+        m_hasClassLevelFailures = true;
       }
     }
 
@@ -555,11 +599,14 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
     // class as failed, or just the current test method as failed, depending on
     // the configuration failure policy
     else if (annotation.getBeforeTestMethod() || annotation.getAfterTestMethod()) {
-      if (m_continueOnFailedConfiguration) {
+      if (m_continueOnFailedConfiguration || canIgnoreConfigFailure(tm)) {
         setMethodInvocationFailure(currentTestMethod, instance);
       } else {
         setClassInvocationFailure(tm.getRealClass(), instance);
       }
+      if (annotation.getBeforeTestMethod()) {
+        m_hasTestMethodLevelFailures = true;
+      }
     }
 
     // If beforeSuite or afterSuite failed, mark *all* the classes as failed
@@ -576,6 +623,9 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
       for (XmlClass xmlClass : classes) {
         setClassInvocationFailure(xmlClass.getSupportClass(), instance);
       }
+      if (annotation.getBeforeTest()) {
+        m_hasTestTagLevelFailures = true;
+      }
     }
     String[] beforeGroups = annotation.getBeforeGroups();
     for (String group : beforeGroups) {
@@ -616,4 +666,44 @@ class ConfigInvoker extends BaseInvoker implements IConfigInvoker {
     }
     return set;
   }
+
+  private static boolean canIgnoreConfigFailure(ITestNGMethod method) {
+    if (method == null) {
+      return false;
+    }
+    return method.isIgnoreFailure();
+  }
+
+  private boolean canIgnoreConfigFailure(IClass testClass, ITestNGMethod configMethod) {
+    boolean instanceMatch = testClass instanceof ITestClass;
+    if (!instanceMatch) {
+      return false;
+    }
+    ITestClass tc = ((ITestClass) testClass);
+    ITestNGMethod[] methods = new ITestNGMethod[] {};
+    if (configMethod == null) { // We are dealing with a test method that is doing the checking
+      // First check if there were any @BeforeMethods that had the isIgnoreFailure flag.
+      if (m_hasTestMethodLevelFailures) {
+        methods = tc.getBeforeTestMethods();
+      } else if (m_hasClassLevelFailures) {
+        // If no @BeforeMethod were found, then we move to @BeforeClass
+        methods = tc.getBeforeClassMethods();
+      } else if (m_hasTestTagLevelFailures) {
+        // If no @BeforeClass were found, then we move to @BeforeTest
+        methods = tc.getBeforeTestConfigurationMethods();
+      }
+    } else { // We are dealing with a config method that is doing the checking
+      // First check if there were any @BeforeTest that had the isIgnoreFailure flag.
+      if (m_hasTestTagLevelFailures) {
+        methods = tc.getBeforeTestConfigurationMethods();
+      } else if (m_hasClassLevelFailures) {
+        // if no @BeforeTest were found, then we move to @BeforeClass
+        methods = tc.getBeforeClassMethods();
+      } else if (m_hasTestMethodLevelFailures) {
+        // if no @BeforeClass were found, then we move to @BeforeMethod
+        methods = tc.getBeforeTestMethods();
+      }
+    }
+    return Arrays.stream(methods).anyMatch(ITestNGMethod::isIgnoreFailure);
+  }
 }
