@@ -21,14 +21,15 @@ import java.util.concurrent.atomic.AtomicInteger;
 import java.util.function.Consumer;
 import java.util.stream.Collectors;
 import org.testng.DataProviderHolder;
-import org.testng.DataProviderInvocationException;
 import org.testng.IClassListener;
 import org.testng.IDataProviderListener;
+import org.testng.IDataProviderMethod;
 import org.testng.IHookable;
 import org.testng.IInvokedMethod;
 import org.testng.IInvokedMethodListener;
 import org.testng.IRetryAnalyzer;
 import org.testng.ISuite;
+import org.testng.ISuiteRunnerListener;
 import org.testng.ITestClass;
 import org.testng.ITestContext;
 import org.testng.ITestListener;
@@ -40,11 +41,13 @@ import org.testng.SuiteRunState;
 import org.testng.SuiteRunner;
 import org.testng.TestException;
 import org.testng.TestNGException;
+import org.testng.TestNotInvokedException;
 import org.testng.collections.CollectionUtils;
 import org.testng.collections.Lists;
 import org.testng.collections.Maps;
 import org.testng.collections.Sets;
 import org.testng.internal.*;
+import org.testng.internal.IObject;
 import org.testng.internal.invokers.GroupConfigMethodArguments.Builder;
 import org.testng.internal.invokers.InvokeMethodRunnable.TestNGRuntimeException;
 import org.testng.internal.thread.ThreadExecutionException;
@@ -68,8 +71,15 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
       DataProviderHolder holder,
       List<IClassListener> m_classListeners,
       boolean m_skipFailedInvocationCounts,
-      ConfigInvoker invoker) {
-    super(m_notifier, m_invokedMethodListeners, m_testContext, m_suiteState, m_configuration);
+      ConfigInvoker invoker,
+      ISuiteRunnerListener suiteRunner) {
+    super(
+        m_notifier,
+        m_invokedMethodListeners,
+        m_testContext,
+        m_suiteState,
+        m_configuration,
+        suiteRunner);
     this.holder = holder;
     this.m_classListeners = m_classListeners;
     this.m_skipFailedInvocationCounts = m_skipFailedInvocationCounts;
@@ -173,9 +183,9 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
 
     ITestClass testClass = testMethod.getTestClass();
     ITestNGMethod[] beforeMethods =
-        TestNgMethodUtils.filterBeforeTestMethods(testClass, CAN_RUN_FROM_CLASS);
+        TestNgMethodUtils.filterBeforeTestMethods(instance, testClass, CAN_RUN_FROM_CLASS);
     ITestNGMethod[] afterMethods =
-        TestNgMethodUtils.filterAfterTestMethods(testClass, CAN_RUN_FROM_CLASS);
+        TestNgMethodUtils.filterAfterTestMethods(instance, testClass, CAN_RUN_FROM_CLASS);
     int invocationCount = onlyOne ? 1 : testMethod.getInvocationCount();
 
     TestMethodArguments arguments =
@@ -226,37 +236,34 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
       int failureCount,
       ITestContext testContext) {
     FailureContext failure = new FailureContext();
-    failure.count = failureCount;
-    failure.representsRetriedMethod = true;
+    failure.count.set(failureCount);
+    failure.representsRetriedMethod.set(true);
     do {
       failure.instances = Lists.newArrayList();
-      Map<String, String> allParameters = Maps.newHashMap();
-      int verbose = testContext.getCurrentXmlTest().getVerbose();
-      // TODO: This recreates all the parameters every time when we only need
-      // one specific set. Should optimize it by only recreating the set needed.
-      ParameterHandler handler =
-          new ParameterHandler(
-              m_configuration.getObjectFactory(), annotationFinder(), this.holder, verbose);
+      boolean cacheData =
+          Optional.ofNullable(arguments.getTestMethod().getDataProviderMethod())
+              .map(IDataProviderMethod::cacheDataForTestRetries)
+              .orElse(false);
+      if (!cacheData) {
+        Map<String, String> allParameters = Maps.newHashMap();
+        int verbose = testContext.getCurrentXmlTest().getVerbose();
+        ParameterHandler handler =
+            new ParameterHandler(
+                m_configuration.getObjectFactory(), annotationFinder(), this.holder, verbose);
 
-      ParameterBag bag =
-          handler.createParameters(
-              arguments.getTestMethod(), arguments.getParameters(), allParameters, testContext);
-      ITestResult errorResult = bag.errorResult;
-
-      if (errorResult != null) {
-        Throwable cause = errorResult.getThrowable();
-        String m = errorResult.getMethod().getMethodName();
-        String msg =
-            String.format(
-                "Encountered problems when gathering parameter values for [%s]. Root cause: ", m);
-        throw new DataProviderInvocationException(msg, cause);
+        ParameterBag bag =
+            handler.createParameters(
+                arguments.getTestMethod(), arguments.getParameters(), allParameters, testContext);
+        if (bag.hasErrors()) {
+          continue;
+        }
       }
       Object[] parameterValues = arguments.getParameterValues();
       TestMethodArguments tma =
           new TestMethodArguments.Builder()
               .usingArguments(arguments)
               .withParameterValues(parameterValues)
-              .withParameters(allParameters)
+              .withParameters(arguments.getParameters())
               .build();
 
       result.add(invokeMethod(tma, testContext.getSuite().getXmlSuite(), failure));
@@ -271,9 +278,13 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
     boolean isFinished = tr.getStatus() != ITestResult.STARTED;
     List<ITestListener> listeners =
         isFinished
-            ? Lists.newReversedArrayList(m_notifier.getTestListeners())
-            : m_notifier.getTestListeners();
+            ? ListenerOrderDeterminer.reversedOrder(
+                m_notifier.getTestListeners(), m_configuration.getListenerComparator())
+            : ListenerOrderDeterminer.order(
+                m_notifier.getTestListeners(), m_configuration.getListenerComparator());
     TestListenerHelper.runTestListeners(tr, listeners);
+    TestListenerHelper.runTestListeners(
+        tr, Collections.singletonList(m_notifier.getExitCodeListener()));
   }
 
   private Collection<IDataProviderListener> dataProviderListeners() {
@@ -288,7 +299,7 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
   }
 
   private DataProviderHolder buildDataProviderHolder() {
-    DataProviderHolder holder = new DataProviderHolder();
+    DataProviderHolder holder = new DataProviderHolder(m_configuration);
     holder.addListeners(dataProviderListeners());
     holder.addInterceptors(this.holder.getInterceptors());
     return holder;
@@ -371,14 +382,14 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
     // Invoke @BeforeGroups on the original method (reduce thread contention,
     // and also solve thread confinement)
     ITestClass testClass = testMethod.getTestClass();
-    Object[] instances = testClass.getInstances(true);
-    for (Object instance : instances) {
+    IObject.IdentifiableObject[] instances = IObject.objects(testClass, true);
+    for (IObject.IdentifiableObject instance : instances) {
       GroupConfigMethodArguments arguments =
           new GroupConfigMethodArguments.Builder()
               .forTestMethod(testMethod)
               .withGroupConfigMethods(groupMethods)
               .withParameters(parameters)
-              .forInstance(instance)
+              .forInstance(instance.getInstance())
               .build();
       invoker.invokeBeforeGroupsConfigurations(arguments);
     }
@@ -386,7 +397,7 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
     long maxTimeOut =
         workers.parallelStream().map(IWorker::getTimeOut).max(Long::compare).orElse(-1L);
 
-    ThreadUtil.execute("methods", workers, threadPoolSize, maxTimeOut);
+    ThreadUtil.execute(m_configuration, "methods", workers, threadPoolSize, maxTimeOut);
 
     //
     // Collect all the TestResults
@@ -544,9 +555,10 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
     } else {
       testResult.setStatus(holder.status);
       if (holder.status == ITestResult.FAILURE && !holder.handled) {
-        int count = failure.count++;
+        int count = failure.count.getAndIncrement();
         if (testMethod.isDataDriven()) {
-          count = 0;
+          String key = Arrays.toString(testResult.getParameters());
+          count = failure.counter.computeIfAbsent(key, k -> new AtomicInteger()).incrementAndGet();
         }
         handleException(testResult.getThrowable(), testMethod, testResult, count);
       }
@@ -591,7 +603,7 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
     long startTime = System.currentTimeMillis();
     InvokedMethod invokedMethod = new InvokedMethod(startTime, testResult);
 
-    if (!failureContext.representsRetriedMethod
+    if (!failureContext.representsRetriedMethod.get()
         && invoker.hasConfigurationFailureFor(
             arguments.getTestMethod(),
             arguments.getTestMethod().getGroups(),
@@ -606,10 +618,9 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
       TestResult.copyAttributes(testResult, result);
       m_notifier.addSkippedTest(arguments.getTestMethod(), result);
       arguments.getTestMethod().incrementCurrentInvocationCount();
-      testResult.setMethod(arguments.getTestMethod());
       invokedMethod = new InvokedMethod(startTime, result);
       invokeListenersForSkippedTestResult(result, invokedMethod);
-      runAfterConfigurations(arguments, suite, testResult);
+      runAfterConfigurations(arguments, suite, result);
       runAfterGroupsConfigurations(arguments);
 
       return result;
@@ -660,28 +671,45 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
               ? (IHookable) arguments.getInstance()
               : m_configuration.getHookable();
 
+      boolean willfullyIgnored = false;
+      boolean usesHookableInstance = hookableInstance != null;
       if (MethodHelper.calculateTimeOut(arguments.getTestMethod()) <= 0) {
-        if (hookableInstance != null) {
-          MethodInvocationHelper.invokeHookable(
-              arguments.getInstance(),
-              arguments.getParameterValues(),
-              hookableInstance,
-              thisMethod,
-              testResult);
+        if (usesHookableInstance) {
+          willfullyIgnored =
+              !MethodInvocationHelper.invokeHookable(
+                  arguments.getInstance(),
+                  arguments.getParameterValues(),
+                  hookableInstance,
+                  thisMethod,
+                  testResult);
         } else {
           // Not a IHookable, invoke directly
           MethodInvocationHelper.invokeMethod(
               thisMethod, arguments.getInstance(), arguments.getParameterValues());
         }
-        setTestStatus(testResult, ITestResult.SUCCESS);
+        if (!willfullyIgnored) {
+          setTestStatus(testResult, ITestResult.SUCCESS);
+        }
       } else {
         // Method with a timeout
-        MethodInvocationHelper.invokeWithTimeout(
-            arguments.getTestMethod(),
-            arguments.getInstance(),
-            arguments.getParameterValues(),
-            testResult,
-            hookableInstance);
+        willfullyIgnored =
+            !MethodInvocationHelper.invokeWithTimeout(
+                m_configuration,
+                arguments.getTestMethod(),
+                arguments.getInstance(),
+                arguments.getParameterValues(),
+                testResult,
+                hookableInstance);
+      }
+      boolean testStatusRemainedUnchanged = testResult.isNotRunning();
+      boolean throwException = !RuntimeBehavior.ignoreCallbackInvocationSkips();
+      if (throwException
+          && usesHookableInstance
+          && willfullyIgnored
+          && testStatusRemainedUnchanged) {
+        TestNotInvokedException tn = new TestNotInvokedException(arguments.tm);
+        testResult.setThrowable(tn);
+        setTestStatus(testResult, ITestResult.FAILURE);
       }
     } catch (InvocationTargetException ite) {
       testResult.setThrowable(ite.getCause());
@@ -713,6 +741,9 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
       StatusHolder holder =
           considerExceptions(
               arguments.getTestMethod(), testResult, expectedExceptionClasses, failureContext);
+      // After considering exceptions, the test status may have gotten updated.
+      // So lets update our test status with the latest status obtained from StatusHolder
+      testResult.setStatus(holder.status);
       runInvokedMethodListeners(AFTER_INVOCATION, invokedMethod, testResult);
       updateStatusHolderAccordingToTestResult(testResult, holder);
       boolean willRetryMethod =
@@ -733,7 +764,9 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
         if (null != testResult.getMethod().getFactoryMethodParamsInfo()) {
           parametersIndex = testResult.getMethod().getFactoryMethodParamsInfo().getIndex();
         }
-        arguments.getTestMethod().addFailedInvocationNumber(parametersIndex);
+        if (!willRetryMethod) {
+          arguments.getTestMethod().addFailedInvocationNumber(parametersIndex);
+        }
       }
 
       //
@@ -765,7 +798,7 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
   }
 
   private void runAfterConfigurations(
-      TestMethodArguments arguments, XmlSuite suite, TestResult testResult) {
+      TestMethodArguments arguments, XmlSuite suite, ITestResult testResult) {
     ITestNGMethod[] teardownConfigMethods =
         TestNgMethodUtils.filterTeardownConfigurationMethods(
             arguments.getTestMethod(), arguments.getAfterMethods());
@@ -787,7 +820,7 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
   private void runConfigMethods(
       TestMethodArguments arguments,
       XmlSuite suite,
-      TestResult testResult,
+      ITestResult testResult,
       ITestNGMethod[] teardownConfigMethods) {
     ConfigMethodArguments cfgArgs =
         new ConfigMethodArguments.Builder()
@@ -819,11 +852,23 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
     return result;
   }
 
+  private static final AutoCloseableLock internalLock = new AutoCloseableLock();
+
   private StatusHolder considerExceptions(
       ITestNGMethod tm,
       ITestResult testResult,
       ExpectedExceptionsHolder exceptionsHolder,
       FailureContext failure) {
+    try (AutoCloseableLock ignore = internalLock.lock()) {
+      return considerExceptionsInternal(tm, testResult, exceptionsHolder, failure);
+    }
+  }
+
+  private StatusHolder considerExceptionsInternal(
+      ITestNGMethod tm,
+      ITestResult testResult,
+      ExpectedExceptionsHolder exceptionsHolder,
+      FailureContext failure) {
     StatusHolder holder = new StatusHolder();
     int status = testResult.getStatus();
     holder.handled = false;
@@ -845,7 +890,7 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
           }
         }
       } else {
-        handleException(ite, tm, testResult, failure.count++);
+        handleException(ite, tm, testResult, failure.count.getAndIncrement());
         holder.handled = true;
         status = testResult.getStatus();
       }
@@ -911,7 +956,12 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
       if (bag.hasErrors()) {
         ITestResult tr = bag.errorResult;
         Throwable throwable = Objects.requireNonNull(tr).getThrowable();
-        if (throwable instanceof TestNGException) {
+        boolean bubbleUpFailures =
+            m_configuration.isPropagateDataProviderFailureAsTestFailure()
+                || bag.isBubbleUpFailures();
+
+        if (!(throwable instanceof SkipException)
+            && (throwable instanceof TestNGException || bubbleUpFailures)) {
           tr.setStatus(ITestResult.FAILURE);
           m_notifier.addFailedTest(arguments.getTestMethod(), tr);
         } else {
@@ -956,6 +1006,8 @@ class TestInvoker extends BaseInvoker implements ITestInvoker {
         ITestResult r =
             TestResult.newEndTimeAwareTestResult(
                 arguments.getTestMethod(), m_testContext, cause, start);
+        Optional.ofNullable(Reporter.getCurrentTestResult())
+            .ifPresent(it -> TestResult.copyAttributes(it, r));
         r.setStatus(TestResult.FAILURE);
         result.add(r);
         runTestResultListener(r);
