@@ -5,7 +5,6 @@ import static org.testng.internal.Utils.isStringBlank;
 import java.io.File;
 import java.io.IOException;
 import java.io.InputStream;
-import java.io.UnsupportedEncodingException;
 import java.net.HttpURLConnection;
 import java.net.MalformedURLException;
 import java.net.URI;
@@ -18,6 +17,7 @@ import java.util.ArrayList;
 import java.util.List;
 import java.util.Map;
 import java.util.Objects;
+import java.util.Optional;
 import java.util.Stack;
 import org.testng.ITestObjectFactory;
 import org.testng.TestNGException;
@@ -101,6 +101,7 @@ public class TestNGContentHandler extends DefaultHandler {
   }
 
   private final Stack<Location> m_locations = new Stack<>();
+  private boolean isSuiteFileTag = false;
 
   private XmlClass m_currentClass = null;
   private ArrayList<XmlInclude> m_currentIncludedMethods = null;
@@ -158,7 +159,7 @@ public class TestNGContentHandler extends DefaultHandler {
   private static boolean isMalformedFileSystemBasedSystemId(String systemId) {
     try {
 
-      URL url = new URL(URLDecoder.decode(systemId, StandardCharsets.UTF_8.name()).trim());
+      URL url = new URL(URLDecoder.decode(systemId, StandardCharsets.UTF_8).trim());
       if (url.getProtocol().equals("file")) {
         File file = new File(url.getFile());
         boolean isDirectory = file.isDirectory();
@@ -166,7 +167,7 @@ public class TestNGContentHandler extends DefaultHandler {
         return isDirectory || !fileExists;
       }
       return false;
-    } catch (MalformedURLException | UnsupportedEncodingException e) {
+    } catch (MalformedURLException e) {
       return true;
     }
   }
@@ -196,9 +197,11 @@ public class TestNGContentHandler extends DefaultHandler {
       String path = attributes.getValue("path");
       pushLocation(Location.SUITE);
       m_suiteFiles.add(path);
+      isSuiteFileTag = true;
     } else {
       m_currentSuite.setSuiteFiles(m_suiteFiles);
       popLocation();
+      isSuiteFileTag = false;
     }
   }
 
@@ -221,7 +224,7 @@ public class TestNGContentHandler extends DefaultHandler {
       }
       String jUnit = attributes.getValue("junit");
       if (null != jUnit) {
-        m_currentSuite.setJUnit(Boolean.valueOf(jUnit));
+        warnSinceJUnitDetected();
       }
       String parallel = attributes.getValue("parallel");
       if (parallel != null) {
@@ -264,6 +267,22 @@ public class TestNGContentHandler extends DefaultHandler {
       if (null != dataProviderThreadCount) {
         m_currentSuite.setDataProviderThreadCount(Integer.parseInt(dataProviderThreadCount));
       }
+
+      String shareThreadPoolForDataProviders =
+          attributes.getValue("share-thread-pool-for-data-providers");
+      Optional.ofNullable(shareThreadPoolForDataProviders)
+          .ifPresent(
+              it ->
+                  m_currentSuite.setShareThreadPoolForDataProviders(
+                      Boolean.parseBoolean(shareThreadPoolForDataProviders)));
+
+      String useGlobalThreadPool = attributes.getValue("use-global-thread-pool");
+      Optional.ofNullable(useGlobalThreadPool)
+          .ifPresent(
+              it ->
+                  m_currentSuite.shouldUseGlobalThreadPool(
+                      Boolean.parseBoolean(useGlobalThreadPool)));
+
       String timeOut = attributes.getValue("time-out");
       if (null != timeOut) {
         m_currentSuite.setTimeOut(timeOut);
@@ -344,7 +363,7 @@ public class TestNGContentHandler extends DefaultHandler {
       }
       String jUnit = attributes.getValue("junit");
       if (null != jUnit) {
-        m_currentTest.setJUnit(Boolean.parseBoolean(jUnit));
+        warnSinceJUnitDetected();
       }
       String skip = attributes.getValue("skipfailedinvocationcounts");
       if (skip != null) {
@@ -388,7 +407,7 @@ public class TestNGContentHandler extends DefaultHandler {
         m_enabledTest = Boolean.parseBoolean(enabledTestString);
       }
     } else {
-      if (null != m_currentTestParameters && m_currentTestParameters.size() > 0) {
+      if (null != m_currentTestParameters && !m_currentTestParameters.isEmpty()) {
         m_currentTest.setParameters(m_currentTestParameters);
       }
       if (null != m_currentClasses) {
@@ -611,6 +630,13 @@ public class TestNGContentHandler extends DefaultHandler {
     } else if ("exclude".equals(qName)) {
       xmlExclude(true, attributes);
     } else if ("parameter".equals(qName)) {
+      if (isSuiteFileTag) {
+        // do-not process a <parameter> tag when it is specified inside <suite-files>
+        Logger.getLogger(getClass())
+            .warn(
+                "Ignoring the <parameter> tag because it is specified inside a <suite-file> tag.");
+        return;
+      }
       String value = expandValue(attributes.getValue("value"));
       Location location = m_locations.peek();
       switch (location) {
@@ -816,4 +842,14 @@ public class TestNGContentHandler extends DefaultHandler {
       return value;
     }
   }
+
+  private static void warnSinceJUnitDetected() {
+    String msg =
+        "Ability to run JUnit tests via TestNG was deprecated in v7.7.0 and is NOW being removed permanently."
+            + " If you would like to run a mixture of TestNG and JUnit tests then please take a look at "
+            + " https://github.com/junit-team/testng-engine which is now maintained by the JUnit team";
+    // Intentionally logging this to the error console so that it's visible to the user.
+    System.err.println(msg);
+    ;
+  }
 }
