@@ -1,5 +1,7 @@
 package test.methods;
 
+import static org.testng.Assert.assertTrue;
+
 import org.testng.annotations.Test;
 
 /**
@@ -41,7 +43,8 @@ public class SampleMethod1 {
   }
 
   public static void verify() {
-    assert m_ok1 && m_ok2 && m_ok3 && m_ok4
-        : "All booleans should be true: " + m_ok1 + " " + m_ok2 + " " + m_ok3 + " " + m_ok4;
+    assertTrue(
+        m_ok1 && m_ok2 && m_ok3 && m_ok4,
+        "All booleans should be true: " + m_ok1 + " " + m_ok2 + " " + m_ok3 + " " + m_ok4);
   }
 }
